#! /usr/bin/env python3

import os
import subprocess
import sys
import platform
import argparse
import inspect
import re
from glob import glob
import multiprocessing
from multiprocessing.pool import ThreadPool
import threading
import tempfile

# See stackoverflow.com/questions/2632199: __file__ nor sys.argv[0]
# are guaranteed to always work, this one should though.
BASEPATH = os.path.dirname(os.path.abspath(inspect.getsourcefile(lambda: None)))


def base_path(*p):
    return os.path.abspath(os.path.join(BASEPATH, *p)).replace("\\", "/")


# Tests require at least CPython 3.3. If your default python3 executable
# is of lower version, you can point MICROPY_CPYTHON3 environment var
# to the correct executable.
if os.name == "nt":
    CPYTHON3 = os.getenv("MICROPY_CPYTHON3", "python")
    MICROPYTHON = os.getenv("MICROPY_MICROPYTHON", base_path("../ports/windows/micropython.exe"))
else:
    CPYTHON3 = os.getenv("MICROPY_CPYTHON3", "python3")
    MICROPYTHON = os.getenv("MICROPY_MICROPYTHON", base_path("../ports/unix/micropython"))

# Use CPython options to not save .pyc files, to only access the core standard library
# (not site packages which may clash with u-module names), and improve start up time.
CPYTHON3_CMD = [CPYTHON3, "-Wignore", "-BS"]

# mpy-cross is only needed if --via-mpy command-line arg is passed
MPYCROSS = os.getenv("MICROPY_MPYCROSS", base_path("../mpy-cross/mpy-cross"))

# For diff'ing test output
DIFF = os.getenv("MICROPY_DIFF", "diff -u")

# Set PYTHONIOENCODING so that CPython will use utf-8 on systems which set another encoding in the locale
os.environ["PYTHONIOENCODING"] = "utf-8"


def rm_f(fname):
    if os.path.exists(fname):
        os.remove(fname)


# unescape wanted regex chars and escape unwanted ones
def convert_regex_escapes(line):
    cs = []
    escape = False
    for c in str(line, "utf8"):
        if escape:
            escape = False
            cs.append(c)
        elif c == "\\":
            escape = True
        elif c in ("(", ")", "[", "]", "{", "}", ".", "*", "+", "^", "$"):
            cs.append("\\" + c)
        else:
            cs.append(c)
    # accept carriage-return(s) before final newline
    if cs[-1] == "\n":
        cs[-1] = "\r*\n"
    return bytes("".join(cs), "utf8")


def run_micropython(pyb, args, test_file, is_special=False):
    special_tests = (
        "micropython/meminfo.py",
        "basics/bytes_compare3.py",
        "basics/builtin_help.py",
        "thread/thread_exc2.py",
        "esp32/partition_ota.py",
    )
    had_crash = False
    if pyb is None:
        # run on PC
        if (
            test_file.startswith(("cmdline/", base_path("feature_check/")))
            or test_file in special_tests
        ):
            # special handling for tests of the unix cmdline program
            is_special = True

        if is_special:
            # check for any cmdline options needed for this test
            args = [MICROPYTHON]
            with open(test_file, "rb") as f:
                line = f.readline()
                if line.startswith(b"# cmdline:"):
                    # subprocess.check_output on Windows only accepts strings, not bytes
                    args += [str(c, "utf-8") for c in line[10:].strip().split()]

            # run the test, possibly with redirected input
            try:
                if "repl_" in test_file:
                    # Need to use a PTY to test command line editing
                    try:
                        import pty
                    except ImportError:
                        # in case pty module is not available, like on Windows
                        return b"SKIP\n"
                    import select

                    def get(required=False):
                        rv = b""
                        while True:
                            ready = select.select([emulator], [], [], 0.02)
                            if ready[0] == [emulator]:
                                rv += os.read(emulator, 1024)
                            else:
                                if not required or rv:
                                    return rv

                    def send_get(what):
                        os.write(emulator, what)
                        return get()

                    with open(test_file, "rb") as f:
                        # instead of: output_mupy = subprocess.check_output(args, stdin=f)
                        # openpty returns two read/write file descriptors.  The first one is
                        # used by the program which provides the virtual
                        # terminal service, and the second one is used by the
                        # subprogram which requires a tty to work.
                        emulator, subterminal = pty.openpty()
                        p = subprocess.Popen(
                            args,
                            stdin=subterminal,
                            stdout=subterminal,
                            stderr=subprocess.STDOUT,
                            bufsize=0,
                        )
                        banner = get(True)
                        output_mupy = banner + b"".join(send_get(line) for line in f)
                        send_get(b"\x04")  # exit the REPL, so coverage info is saved
                        # At this point the process might have exited already, but trying to
                        # kill it 'again' normally doesn't result in exceptions as Python and/or
                        # the OS seem to try to handle this nicely. When running Linux on WSL
                        # though, the situation differs and calling Popen.kill after the process
                        # terminated results in a ProcessLookupError. Just catch that one here
                        # since we just want the process to be gone and that's the case.
                        try:
                            p.kill()
                        except ProcessLookupError:
                            pass
                        os.close(emulator)
                        os.close(subterminal)
                else:
                    output_mupy = subprocess.check_output(
                        args + [test_file], stderr=subprocess.STDOUT
                    )

            except subprocess.CalledProcessError as error:
                return error.output + b"CRASH"

        else:
            # a standard test run on PC

            # create system command
            cmdlist = [MICROPYTHON, "-X", "emit=" + args.emit]
            if args.heapsize is not None:
                cmdlist.extend(["-X", "heapsize=" + args.heapsize])

            # if running via .mpy, first compile the .py file
            if args.via_mpy:
                mpy_modname = tempfile.mktemp(dir="")
                mpy_filename = mpy_modname + ".mpy"
                subprocess.check_output(
                    [MPYCROSS]
                    + args.mpy_cross_flags.split()
                    + ["-o", mpy_filename, "-X", "emit=" + args.emit, test_file]
                )
                cmdlist.extend(["-m", mpy_modname])
            else:
                cmdlist.append(test_file)

            # run the actual test
            try:
                output_mupy = subprocess.check_output(cmdlist, stderr=subprocess.STDOUT)
            except subprocess.CalledProcessError as er:
                had_crash = True
                output_mupy = er.output + b"CRASH"

            # clean up if we had an intermediate .mpy file
            if args.via_mpy:
                rm_f(mpy_filename)

    else:
        # run on pyboard
        pyb.enter_raw_repl()
        try:
            output_mupy = pyb.execfile(test_file)
        except pyboard.PyboardError as e:
            had_crash = True
            if not is_special and e.args[0] == "exception":
                output_mupy = e.args[1] + e.args[2] + b"CRASH"
            else:
                output_mupy = bytes(e.args[0], "ascii") + b"\nCRASH"

    # canonical form for all ports/platforms is to use \n for end-of-line
    output_mupy = output_mupy.replace(b"\r\n", b"\n")

    # don't try to convert the output if we should skip this test
    if had_crash or output_mupy in (b"SKIP\n", b"CRASH"):
        return output_mupy

    if is_special or test_file in special_tests:
        # convert parts of the output that are not stable across runs
        with open(test_file + ".exp", "rb") as f:
            lines_exp = []
            for line in f.readlines():
                if line == b"########\n":
                    line = (line,)
                else:
                    line = (line, re.compile(convert_regex_escapes(line)))
                lines_exp.append(line)
        lines_mupy = [line + b"\n" for line in output_mupy.split(b"\n")]
        if output_mupy.endswith(b"\n"):
            lines_mupy = lines_mupy[:-1]  # remove erroneous last empty line
        i_mupy = 0
        for i in range(len(lines_exp)):
            if lines_exp[i][0] == b"########\n":
                # 8x #'s means match 0 or more whole lines
                line_exp = lines_exp[i + 1]
                skip = 0
                while i_mupy + skip < len(lines_mupy) and not line_exp[1].match(
                    lines_mupy[i_mupy + skip]
                ):
                    skip += 1
                if i_mupy + skip >= len(lines_mupy):
                    lines_mupy[i_mupy] = b"######## FAIL\n"
                    break
                del lines_mupy[i_mupy : i_mupy + skip]
                lines_mupy.insert(i_mupy, b"########\n")
                i_mupy += 1
            else:
                # a regex
                if lines_exp[i][1].match(lines_mupy[i_mupy]):
                    lines_mupy[i_mupy] = lines_exp[i][0]
                else:
                    # print("don't match: %r %s" % (lines_exp[i][1], lines_mupy[i_mupy])) # DEBUG
                    pass
                i_mupy += 1
            if i_mupy >= len(lines_mupy):
                break
        output_mupy = b"".join(lines_mupy)

    return output_mupy


def run_feature_check(pyb, args, base_path, test_file):
    if pyb is not None and test_file.startswith("repl_"):
        # REPL feature tests will not run via pyboard because they require prompt interactivity
        return b""
    return run_micropython(pyb, args, base_path("feature_check", test_file), is_special=True)


class ThreadSafeCounter:
    def __init__(self, start=0):
        self._value = start
        self._lock = threading.Lock()

    def increment(self):
        self.add(1)

    def add(self, to_add):
        with self._lock:
            self._value += to_add

    def append(self, arg):
        self.add([arg])

    @property
    def value(self):
        return self._value


def run_tests(pyb, tests, args, result_dir, num_threads=1):
    test_count = ThreadSafeCounter()
    testcase_count = ThreadSafeCounter()
    passed_count = ThreadSafeCounter()
    failed_tests = ThreadSafeCounter([])
    skipped_tests = ThreadSafeCounter([])

    skip_tests = set()
    skip_native = False
    skip_int_big = False
    skip_bytearray = False
    skip_set_type = False
    skip_slice = False
    skip_async = False
    skip_const = False
    skip_revops = False
    skip_io_module = False
    skip_endian = False
    has_complex = True
    has_coverage = False

    upy_float_precision = 32

    # If we're asked to --list-tests, we can't assume that there's a
    # connection to target, so we can't run feature checks usefully.
    if not (args.list_tests or args.write_exp):
        # Even if we run completely different tests in a different directory,
        # we need to access feature_checks from the same directory as the
        # run-tests.py script itself so use base_path.

        # Check if micropython.native is supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "native_check.py")
        if output != b"native\n":
            skip_native = True

        # Check if arbitrary-precision integers are supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "int_big.py")
        if output != b"1000000000000000000000000000000000000000000000\n":
            skip_int_big = True

        # Check if bytearray is supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "bytearray.py")
        if output != b"bytearray\n":
            skip_bytearray = True

        # Check if set type (and set literals) is supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "set_check.py")
        if output != b"{1}\n":
            skip_set_type = True

        # Check if slice is supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "slice.py")
        if output != b"slice\n":
            skip_slice = True

        # Check if async/await keywords are supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "async_check.py")
        if output != b"async\n":
            skip_async = True

        # Check if const keyword (MicroPython extension) is supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "const.py")
        if output != b"1\n":
            skip_const = True

        # Check if __rOP__ special methods are supported, and skip such tests if it's not
        output = run_feature_check(pyb, args, base_path, "reverse_ops.py")
        if output == b"TypeError\n":
            skip_revops = True

        # Check if uio module exists, and skip such tests if it doesn't
        output = run_feature_check(pyb, args, base_path, "uio_module.py")
        if output != b"uio\n":
            skip_io_module = True

        # Check if emacs repl is supported, and skip such tests if it's not
        t = run_feature_check(pyb, args, base_path, "repl_emacs_check.py")
        if "True" not in str(t, "ascii"):
            skip_tests.add("cmdline/repl_emacs_keys.py")

        # Check if words movement in repl is supported, and skip such tests if it's not
        t = run_feature_check(pyb, args, base_path, "repl_words_move_check.py")
        if "True" not in str(t, "ascii"):
            skip_tests.add("cmdline/repl_words_move.py")

        upy_byteorder = run_feature_check(pyb, args, base_path, "byteorder.py")
        upy_float_precision = run_feature_check(pyb, args, base_path, "float.py")
        try:
            upy_float_precision = int(upy_float_precision)
        except ValueError:
            upy_float_precision = 0
        has_complex = run_feature_check(pyb, args, base_path, "complex.py") == b"complex\n"
        has_coverage = run_feature_check(pyb, args, base_path, "coverage.py") == b"coverage\n"
        cpy_byteorder = subprocess.check_output(
            CPYTHON3_CMD + [base_path("feature_check/byteorder.py")]
        )
        skip_endian = upy_byteorder != cpy_byteorder

    # These tests don't test slice explicitly but rather use it to perform the test
    misc_slice_tests = (
        "builtin_range",
        "class_super",
        "containment",
        "errno1",
        "fun_str",
        "generator1",
        "globals_del",
        "memoryview1",
        "memoryview_gc",
        "object1",
        "python34",
        "struct_endian",
    )

    # Some tests shouldn't be run on GitHub Actions
    if os.getenv("GITHUB_ACTIONS") == "true":
        skip_tests.add("thread/stress_schedule.py")  # has reliability issues

    if upy_float_precision == 0:
        skip_tests.add("extmod/uctypes_le_float.py")
        skip_tests.add("extmod/uctypes_native_float.py")
        skip_tests.add("extmod/uctypes_sizeof_float.py")
        skip_tests.add("extmod/ujson_dumps_float.py")
        skip_tests.add("extmod/ujson_loads_float.py")
        skip_tests.add("extmod/urandom_extra_float.py")
        skip_tests.add("misc/rge_sm.py")
    if upy_float_precision < 32:
        skip_tests.add(
            "float/float2int_intbig.py"
        )  # requires fp32, there's float2int_fp30_intbig.py instead
        skip_tests.add(
            "float/string_format.py"
        )  # requires fp32, there's string_format_fp30.py instead
        skip_tests.add("float/bytes_construct.py")  # requires fp32
        skip_tests.add("float/bytearray_construct.py")  # requires fp32
    if upy_float_precision < 64:
        skip_tests.add("float/float_divmod.py")  # tested by float/float_divmod_relaxed.py instead
        skip_tests.add("float/float2int_doubleprec_intbig.py")
        skip_tests.add("float/float_parse_doubleprec.py")

    if not has_complex:
        skip_tests.add("float/complex1.py")
        skip_tests.add("float/complex1_intbig.py")
        skip_tests.add("float/complex_special_methods.py")
        skip_tests.add("float/int_big_float.py")
        skip_tests.add("float/true_value.py")
        skip_tests.add("float/types.py")

    if not has_coverage:
        skip_tests.add("cmdline/cmd_parsetree.py")

    # Some tests shouldn't be run on a PC
    if args.target == "unix":
        # unix build does not have the GIL so can't run thread mutation tests
        for t in tests:
            if t.startswith("thread/mutate_"):
                skip_tests.add(t)

    # Some tests shouldn't be run on pyboard
    if args.target != "unix":
        skip_tests.add("basics/exception_chain.py")  # warning is not printed
        skip_tests.add("micropython/meminfo.py")  # output is very different to PC output
        skip_tests.add("extmod/machine_mem.py")  # raw memory access not supported

        if args.target == "wipy":
            skip_tests.add("misc/print_exception.py")  # requires error reporting full
            skip_tests.update(
                {
                    "extmod/uctypes_%s.py" % t
                    for t in "bytearray le native_le ptr_le ptr_native_le sizeof sizeof_native array_assign_le array_assign_native_le".split()
                }
            )  # requires uctypes
            skip_tests.add("extmod/zlibd_decompress.py")  # requires zlib
            skip_tests.add("extmod/uheapq1.py")  # uheapq not supported by WiPy
            skip_tests.add("extmod/urandom_basic.py")  # requires urandom
            skip_tests.add("extmod/urandom_extra.py")  # requires urandom
        elif args.target == "esp8266":
            skip_tests.add("misc/rge_sm.py")  # too large
        elif args.target == "minimal":
            skip_tests.add("basics/class_inplace_op.py")  # all special methods not supported
            skip_tests.add(
                "basics/subclass_native_init.py"
            )  # native subclassing corner cases not support
            skip_tests.add("misc/rge_sm.py")  # too large
            skip_tests.add("micropython/opt_level.py")  # don't assume line numbers are stored
        elif args.target == "nrf":
            skip_tests.add("basics/memoryview1.py")  # no item assignment for memoryview
            skip_tests.add("extmod/urandom_basic.py")  # unimplemented: urandom.seed
            skip_tests.add("micropython/opt_level.py")  # no support for line numbers
            skip_tests.add("misc/non_compliant.py")  # no item assignment for bytearray
            for t in tests:
                if t.startswith("basics/io_"):
                    skip_tests.add(t)
        elif args.target == "qemu-arm":
            skip_tests.add("misc/print_exception.py")  # requires sys stdfiles

    # Some tests are known to fail on 64-bit machines
    if pyb is None and platform.architecture()[0] == "64bit":
        pass

    # Some tests use unsupported features on Windows
    if os.name == "nt":
        skip_tests.add("import/import_file.py")  # works but CPython prints forward slashes

    # Some tests are known to fail with native emitter
    # Remove them from the below when they work
    if args.emit == "native":
        skip_tests.update(
            {"basics/%s.py" % t for t in "gen_yield_from_close generator_name".split()}
        )  # require raise_varargs, generator name
        skip_tests.update(
            {"basics/async_%s.py" % t for t in "with with2 with_break with_return".split()}
        )  # require async_with
        skip_tests.update(
            {"basics/%s.py" % t for t in "try_reraise try_reraise2".split()}
        )  # require raise_varargs
        skip_tests.add("basics/annotate_var.py")  # requires checking for unbound local
        skip_tests.add("basics/del_deref.py")  # requires checking for unbound local
        skip_tests.add("basics/del_local.py")  # requires checking for unbound local
        skip_tests.add("basics/exception_chain.py")  # raise from is not supported
        skip_tests.add("basics/scope_implicit.py")  # requires checking for unbound local
        skip_tests.add("basics/try_finally_return2.py")  # requires raise_varargs
        skip_tests.add("basics/unboundlocal.py")  # requires checking for unbound local
        skip_tests.add("extmod/uasyncio_event.py")  # unknown issue
        skip_tests.add("extmod/uasyncio_lock.py")  # requires async with
        skip_tests.add("extmod/uasyncio_micropython.py")  # unknown issue
        skip_tests.add("extmod/uasyncio_wait_for.py")  # unknown issue
        skip_tests.add("misc/features.py")  # requires raise_varargs
        skip_tests.add(
            "misc/print_exception.py"
        )  # because native doesn't have proper traceback info
        skip_tests.add("misc/sys_exc_info.py")  # sys.exc_info() is not supported for native
        skip_tests.add(
            "micropython/emg_exc.py"
        )  # because native doesn't have proper traceback info
        skip_tests.add(
            "micropython/heapalloc_traceback.py"
        )  # because native doesn't have proper traceback info
        skip_tests.add(
            "micropython/opt_level_lineno.py"
        )  # native doesn't have proper traceback info
        skip_tests.add("micropython/schedule.py")  # native code doesn't check pending events

    def run_one_test(test_file):
        test_file = test_file.replace("\\", "/")

        if args.filters:
            # Default verdict is the opposit of the first action
            verdict = "include" if args.filters[0][0] == "exclude" else "exclude"
            for action, pat in args.filters:
                if pat.search(test_file):
                    verdict = action
            if verdict == "exclude":
                return

        test_basename = test_file.replace("..", "_").replace("./", "").replace("/", "_")
        test_name = os.path.splitext(os.path.basename(test_file))[0]
        is_native = (
            test_name.startswith("native_")
            or test_name.startswith("viper_")
            or args.emit == "native"
        )
        is_endian = test_name.endswith("_endian")
        is_int_big = test_name.startswith("int_big") or test_name.endswith("_intbig")
        is_bytearray = test_name.startswith("bytearray") or test_name.endswith("_bytearray")
        is_set_type = test_name.startswith("set_") or test_name.startswith("frozenset")
        is_slice = test_name.find("slice") != -1 or test_name in misc_slice_tests
        is_async = test_name.startswith(("async_", "uasyncio_"))
        is_const = test_name.startswith("const")
        is_io_module = test_name.startswith("io_")

        skip_it = test_file in skip_tests
        skip_it |= skip_native and is_native
        skip_it |= skip_endian and is_endian
        skip_it |= skip_int_big and is_int_big
        skip_it |= skip_bytearray and is_bytearray
        skip_it |= skip_set_type and is_set_type
        skip_it |= skip_slice and is_slice
        skip_it |= skip_async and is_async
        skip_it |= skip_const and is_const
        skip_it |= skip_revops and "reverse_op" in test_name
        skip_it |= skip_io_module and is_io_module

        if args.list_tests:
            if not skip_it:
                print(test_file)
            return

        if skip_it:
            print("skip ", test_file)
            skipped_tests.append(test_name)
            return

        # get expected output
        test_file_expected = test_file + ".exp"
        if os.path.isfile(test_file_expected):
            # expected output given by a file, so read that in
            with open(test_file_expected, "rb") as f:
                output_expected = f.read()
        else:
            e = {"PYTHONPATH": os.getcwd(), "PATH": os.environ["PATH"], "LANG": "en_US.UTF-8"}
            # run CPython to work out expected output
            try:
                output_expected = subprocess.check_output(
                    CPYTHON3_CMD + [test_file], env=e, stderr=subprocess.STDOUT
                )
                if args.write_exp:
                    with open(test_file_expected, "wb") as f:
                        f.write(output_expected)
            except subprocess.CalledProcessError as error:
                output_expected = error.output + b"CPYTHON3 CRASH"

        # canonical form for all host platforms is to use \n for end-of-line
        output_expected = output_expected.replace(b"\r\n", b"\n")

        if args.write_exp:
            return

        # run MicroPython
        output_mupy = run_micropython(pyb, args, test_file)

        if output_mupy == b"SKIP\n":
            print("skip ", test_file)
            skipped_tests.append(test_name)
            return

        testcase_count.add(len(output_expected.splitlines()))

        filename_expected = os.path.join(result_dir, test_basename + ".exp")
        filename_mupy = os.path.join(result_dir, test_basename + ".out")

        if output_expected == output_mupy:
            print("pass ", test_file)
            passed_count.increment()
            rm_f(filename_expected)
            rm_f(filename_mupy)
        else:
            with open(filename_expected, "wb") as f:
                f.write(output_expected)
            with open(filename_mupy, "wb") as f:
                f.write(output_mupy)
            print("FAIL ", test_file)
            failed_tests.append(test_name)

        test_count.increment()

    if pyb or args.list_tests:
        num_threads = 1

    if num_threads > 1:
        pool = ThreadPool(num_threads)
        pool.map(run_one_test, tests)
    else:
        for test in tests:
            run_one_test(test)

    if args.list_tests:
        return True

    print(
        "{} tests performed ({} individual testcases)".format(
            test_count.value, testcase_count.value
        )
    )
    print("{} tests passed".format(passed_count.value))

    skipped_tests = sorted(skipped_tests.value)
    if len(skipped_tests) > 0:
        print("{} tests skipped: {}".format(len(skipped_tests), " ".join(skipped_tests)))
    failed_tests = sorted(failed_tests.value)
    if len(failed_tests) > 0:
        print("{} tests failed: {}".format(len(failed_tests), " ".join(failed_tests)))
        return False

    # all tests succeeded
    return True


class append_filter(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, default=[], **kwargs)

    def __call__(self, parser, args, value, option):
        if not hasattr(args, self.dest):
            args.filters = []
        if option.startswith(("-e", "--e")):
            option = "exclude"
        else:
            option = "include"
        args.filters.append((option, re.compile(value)))


def main():
    cmd_parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Run and manage tests for MicroPython.

Tests are discovered by scanning test directories for .py files or using the
specified test files. If test files nor directories are specified, the script
expects to be ran in the tests directory (where this file is located) and the
builtin tests suitable for the target platform are ran.
When running tests, run-tests.py compares the MicroPython output of the test with the output
produced by running the test through CPython unless a <test>.exp file is found, in which
case it is used as comparison.
If a test fails, run-tests.py produces a pair of <test>.out and <test>.exp files in the result
directory with the MicroPython output and the expectations, respectively.
""",
        epilog="""\
Options -i and -e can be multiple and processed in the order given. Regex
"search" (vs "match") operation is used. An action (include/exclude) of
the last matching regex is used:
  run-tests.py -i async - exclude all, then include tests containing "async" anywhere
  run-tests.py -e '/big.+int' - include all, then exclude by regex
  run-tests.py -e async -i async_foo - include all, exclude async, yet still include async_foo
""",
    )
    cmd_parser.add_argument("--target", default="unix", help="the target platform")
    cmd_parser.add_argument(
        "--device",
        default="/dev/ttyACM0",
        help="the serial device or the IP address of the pyboard",
    )
    cmd_parser.add_argument(
        "-b", "--baudrate", default=115200, help="the baud rate of the serial device"
    )
    cmd_parser.add_argument("-u", "--user", default="micro", help="the telnet login username")
    cmd_parser.add_argument("-p", "--password", default="python", help="the telnet login password")
    cmd_parser.add_argument(
        "-d", "--test-dirs", nargs="*", help="input test directories (if no files given)"
    )
    cmd_parser.add_argument(
        "-r", "--result-dir", default=base_path("results"), help="directory for test results"
    )
    cmd_parser.add_argument(
        "-e",
        "--exclude",
        action=append_filter,
        metavar="REGEX",
        dest="filters",
        help="exclude test by regex on path/name.py",
    )
    cmd_parser.add_argument(
        "-i",
        "--include",
        action=append_filter,
        metavar="REGEX",
        dest="filters",
        help="include test by regex on path/name.py",
    )
    cmd_parser.add_argument(
        "--write-exp",
        action="store_true",
        help="use CPython to generate .exp files to run tests w/o CPython",
    )
    cmd_parser.add_argument(
        "--list-tests", action="store_true", help="list tests instead of running them"
    )
    cmd_parser.add_argument(
        "--emit", default="bytecode", help="MicroPython emitter to use (bytecode or native)"
    )
    cmd_parser.add_argument("--heapsize", help="heapsize to use (use default if not specified)")
    cmd_parser.add_argument(
        "--via-mpy", action="store_true", help="compile .py files to .mpy first"
    )
    cmd_parser.add_argument(
        "--mpy-cross-flags", default="-mcache-lookup-bc", help="flags to pass to mpy-cross"
    )
    cmd_parser.add_argument(
        "--keep-path", action="store_true", help="do not clear MICROPYPATH when running tests"
    )
    cmd_parser.add_argument(
        "-j",
        "--jobs",
        default=multiprocessing.cpu_count(),
        metavar="N",
        type=int,
        help="Number of tests to run simultaneously",
    )
    cmd_parser.add_argument("files", nargs="*", help="input test files")
    cmd_parser.add_argument(
        "--print-failures",
        action="store_true",
        help="print the diff of expected vs. actual output for failed tests and exit",
    )
    cmd_parser.add_argument(
        "--clean-failures",
        action="store_true",
        help="delete the .exp and .out files from failed tests and exit",
    )
    args = cmd_parser.parse_args()

    if args.print_failures:
        for exp in glob(os.path.join(args.result_dir, "*.exp")):
            testbase = exp[:-4]
            print()
            print("FAILURE {0}".format(testbase))
            os.system("{0} {1}.exp {1}.out".format(DIFF, testbase))

        sys.exit(0)

    if args.clean_failures:
        for f in glob(os.path.join(args.result_dir, "*.exp")) + glob(
            os.path.join(args.result_dir, "*.out")
        ):
            os.remove(f)

        sys.exit(0)

    LOCAL_TARGETS = (
        "unix",
        "qemu-arm",
    )
    EXTERNAL_TARGETS = ("pyboard", "wipy", "esp8266", "esp32", "minimal", "nrf")
    if args.target in LOCAL_TARGETS or args.list_tests:
        pyb = None
    elif args.target in EXTERNAL_TARGETS:
        global pyboard
        sys.path.append(base_path("../tools"))
        import pyboard

        pyb = pyboard.Pyboard(args.device, args.baudrate, args.user, args.password)
        pyb.enter_raw_repl()
    else:
        raise ValueError("target must be one of %s" % ", ".join(LOCAL_TARGETS + EXTERNAL_TARGETS))

    if len(args.files) == 0:
        if args.test_dirs is None:
            test_dirs = (
                "basics",
                "micropython",
                "misc",
                "extmod",
            )
            if args.target == "pyboard":
                # run pyboard tests
                test_dirs += ("float", "stress", "pyb", "pybnative", "inlineasm")
            elif args.target in ("esp8266", "esp32", "minimal", "nrf"):
                test_dirs += ("float",)
            elif args.target == "wipy":
                # run WiPy tests
                test_dirs += ("wipy",)
            elif args.target == "unix":
                # run PC tests
                test_dirs += (
                    "float",
                    "import",
                    "io",
                    "stress",
                    "unicode",
                    "unix",
                    "cmdline",
                    "../extmod/ulab/tests",
                )
            elif args.target == "qemu-arm":
                if not args.write_exp:
                    raise ValueError("--target=qemu-arm must be used with --write-exp")
                # Generate expected output files for qemu run.
                # This list should match the test_dirs tuple in tinytest-codegen.py.
                test_dirs += (
                    "float",
                    "inlineasm",
                    "qemu-arm",
                )
        else:
            # run tests from these directories
            test_dirs = args.test_dirs
        tests = sorted(
            test_file
            for test_files in (glob("{}/*.py".format(dir)) for dir in test_dirs)
            for test_file in test_files
        )
    else:
        # tests explicitly given
        tests = args.files

    if not args.keep_path:
        # clear search path to make sure tests use only builtin modules and those in extmod
        os.environ["MICROPYPATH"] = os.pathsep + base_path("../extmod")

    try:
        os.makedirs(args.result_dir, exist_ok=True)
        res = run_tests(pyb, tests, args, args.result_dir, args.jobs)
    finally:
        if pyb:
            pyb.close()

    if not res:
        sys.exit(1)


if __name__ == "__main__":
    main()



# Derived from code on Eric Holscher's blog, found at:
# https://www.ericholscher.com/blog/2016/jul/25/integrating-jinja-rst-sphinx/

def rstjinja(app, docname, source):
    """
    Render our pages as a jinja template for fancy templating goodness.
    """
    # Make sure we're outputting HTML
    if app.builder.format not in ("html", "latex"):
        return

    # we only want our one jinja template to run through this func
    if "shared-bindings/support_matrix" not in docname:
        return

    src = rendered = source[0]
    print(docname)

    if app.builder.format == "html":
        rendered = app.builder.templates.render_string(
            src, app.config.html_context
        )
    else:
        from sphinx.util.template import BaseRenderer
        renderer = BaseRenderer()
        rendered = renderer.render_string(
            src,
            app.config.html_context
        )

    source[0] = rendered

def setup(app):
    app.connect("source-read", rstjinja)



# Test fairness of cancelling a task
# That tasks which keeps being cancelled by multiple other tasks gets a chance to run

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def task_a():
    try:
        while True:
            print("sleep a")
            await asyncio.sleep(0)
    except asyncio.CancelledError:
        print("cancelled a")


async def task_b(id, other):
    while other.cancel():
        print("sleep b", id)
        await asyncio.sleep(0)
    print("done b", id)


async def main():
    t = asyncio.create_task(task_a())
    for i in range(3):
        asyncio.create_task(task_b(i, t))
    await t


asyncio.run(main())



try:
    from ucryptolib import aes
except ImportError:
    print("SKIP")
    raise SystemExit


def _new(k, ctr_initial):
    return aes(k, 6, ctr_initial)


try:
    _new(b"x" * 16, b"x" * 16)
except ValueError as e:
    # is CTR support disabled?
    if e.args[0] == "mode":
        print("SKIP")
        raise SystemExit
    raise e

crypto = _new(b"1234" * 4, b"5678" * 4)
enc = crypto.encrypt(b"a")
print(enc)
enc += crypto.encrypt(b"b" * 1000)
print(enc)

crypto = _new(b"1234" * 4, b"5678" * 4)
print(crypto.decrypt(enc))



# Test get_event_loop()

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def main():
    print("start")
    await asyncio.sleep(0.01)
    print("end")


loop = asyncio.get_event_loop()
loop.run_until_complete(main())



# set type

# This doesn't really work as expected, because {None}
# leads SyntaxError during parsing.
try:
    set
except NameError:
    print("SKIP")
    raise SystemExit

print(set)

print(type(set()) == set)

print(type({None}) == set)



# test round() with integer values and second arg

# rounding integers is an optional feature so test for it
try:
    round(1, -1)
except NotImplementedError:
    print('SKIP')
    raise SystemExit

tests = [
    (1, False), (1, True),
    (124, -1), (125, -1), (126, -1),
    (5, -1), (15, -1), (25, -1),
    (12345, 0), (12345, -1), (12345, 1),
    (-1234, 0), (-1234, -1), (-1234, 1),
]
for t in tests:
    print(round(*t))



@micropython.asm_thumb  # r0 = r0+r1-r2
def add_sub(r0, r1, r2):
    vmov(s0, r0)
    vcvt_f32_s32(s0, s0)
    vmov(s1, r1)
    vcvt_f32_s32(s1, s1)
    vmov(s2, r2)
    vcvt_f32_s32(s2, s2)
    vadd(s0, s0, s1)
    vsub(s0, s0, s2)
    vcvt_s32_f32(s31, s0)
    vmov(r0, s31)


print(add_sub(100, 20, 30))



# Test VfsFat class and its finaliser

try:
    import uerrno, uos

    uos.VfsFat
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMBlockDevice:
    def __init__(self, blocks, sec_size=512):
        self.sec_size = sec_size
        self.data = bytearray(blocks * self.sec_size)

    def readblocks(self, n, buf):
        for i in range(len(buf)):
            buf[i] = self.data[n * self.sec_size + i]

    def writeblocks(self, n, buf):
        for i in range(len(buf)):
            self.data[n * self.sec_size + i] = buf[i]

    def ioctl(self, op, arg):
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return len(self.data) // self.sec_size
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.sec_size


# Create block device, and skip test if not enough RAM
try:
    bdev = RAMBlockDevice(50)
except MemoryError:
    print("SKIP")
    raise SystemExit

# Format block device and create VFS object
uos.VfsFat.mkfs(bdev)
vfs = uos.VfsFat(bdev)

# Here we test that opening a file with the heap locked fails correctly.  This
# is a special case because file objects use a finaliser and allocating with a
# finaliser is a different path to normal allocation.  It would be better to
# test this in the core tests but there are no core objects that use finaliser.
import micropython

micropython.heap_lock()
try:
    vfs.open("x", "r")
except MemoryError:
    print("MemoryError")
micropython.heap_unlock()

# Here we test that the finaliser is actually called during a garbage collection.
import gc

N = 4
for i in range(N):
    n = "x%d" % i
    f = vfs.open(n, "w")
    f.write(n)
    f = None  # release f without closing
    [0, 1, 2, 3, 4, 5]  # use up Python stack so f is really gone
gc.collect()  # should finalise all N files by closing them
for i in range(N):
    with vfs.open("x%d" % i, "r") as f:
        print(f.read())



try:
    from uio import StringIO
    import ujson as json
except:
    try:
        from io import StringIO
        import json
    except ImportError:
        print("SKIP")
        raise SystemExit

s = StringIO()
json.dump(False, s)
print(s.getvalue())

s = StringIO()
json.dump({"a": (2, [3, None])}, s)
print(s.getvalue())

# dump to a small-int not allowed
try:
    json.dump(123, 1)
except (AttributeError, OSError):  # CPython and uPy have different errors
    print("Exception")

# dump to an object not allowed
try:
    json.dump(123, {})
except (AttributeError, OSError):  # CPython and uPy have different errors
    print("Exception")



# test simple async with execution

class AContext:
    async def __aenter__(self):
        print('enter')
        return 1
    async def __aexit__(self, exc_type, exc, tb):
        print('exit', exc_type, exc)

async def f():
    async with AContext():
        print('body')

o = f()
try:
    o.send(None)
except StopIteration:
    print('finished')

async def g():
    async with AContext() as ac:
        print(ac)
        raise ValueError('error')

o = g()
try:
    o.send(None)
except ValueError:
    print('ValueError')

# test raising BaseException to make sure it is handled by the async-with
async def h():
    async with AContext():
        raise BaseException
o = h()
try:
    o.send(None)
except BaseException:
    print('BaseException')



# builtin type that is not iterable
try:
    for i in 1:
        pass
except TypeError:
    print('TypeError')

# builtin type that is iterable, calling __next__ explicitly
print(iter(range(4)).__next__())



# test concurrent mutating access to a shared user instance
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import _thread

# the shared user class and instance
class User:
    def __init__(self):
        self.a = "A"
        self.b = "B"
        self.c = "C"


user = User()

# main thread function
def th(n, lo, hi):
    for repeat in range(n):
        for i in range(lo, hi):
            setattr(user, "attr_%u" % i, repeat + i)
            assert getattr(user, "attr_%u" % i) == repeat + i

    with lock:
        global n_finished
        n_finished += 1


lock = _thread.allocate_lock()
n_repeat = 30
n_range = 50  # 300 for stressful test (uses more heap)
n_thread = 4
n_finished = 0

# spawn threads
for i in range(n_thread):
    _thread.start_new_thread(th, (n_repeat, i * n_range, (i + 1) * n_range))

# busy wait for threads to finish
while n_finished < n_thread:
    pass

# check user instance has correct contents
print(user.a, user.b, user.c)
for i in range(n_thread * n_range):
    assert getattr(user, "attr_%u" % i) == n_repeat - 1 + i



"""
This script processes the output from the C preprocessor and extracts all
qstr. Each qstr is transformed into a qstr definition of the form 'Q(...)'.

This script works with Python 2.6, 2.7, 3.3 and 3.4.
"""

from __future__ import print_function

import io
import os
import re
import subprocess
import sys
import multiprocessing, multiprocessing.dummy

# Python 2/3 compatibility:
#   - iterating through bytes is different
#   - codepoint2name lives in a different module
import platform

if platform.python_version_tuple()[0] == "2":
    bytes_cons = lambda val, enc=None: bytearray(val)
    from htmlentitydefs import name2codepoint
elif platform.python_version_tuple()[0] == "3":
    bytes_cons = bytes
    from html.entities import name2codepoint

    unichr = chr
# end compatibility code

# Blocklist of qstrings that are specially handled in further
# processing and should be ignored
QSTRING_BLOCK_LIST = set(["NULL", "number_of"])

# add some custom names to map characters that aren't in HTML
name2codepoint["hyphen"] = ord("-")
name2codepoint["space"] = ord(" ")
name2codepoint["squot"] = ord("'")
name2codepoint["comma"] = ord(",")
name2codepoint["dot"] = ord(".")
name2codepoint["colon"] = ord(":")
name2codepoint["semicolon"] = ord(";")
name2codepoint["slash"] = ord("/")
name2codepoint["percent"] = ord("%")
name2codepoint["hash"] = ord("#")
name2codepoint["paren_open"] = ord("(")
name2codepoint["paren_close"] = ord(")")
name2codepoint["bracket_open"] = ord("[")
name2codepoint["bracket_close"] = ord("]")
name2codepoint["brace_open"] = ord("{")
name2codepoint["brace_close"] = ord("}")
name2codepoint["star"] = ord("*")
name2codepoint["bang"] = ord("!")
name2codepoint["backslash"] = ord("\\")
name2codepoint["plus"] = ord("+")
name2codepoint["dollar"] = ord("$")
name2codepoint["equals"] = ord("=")
name2codepoint["question"] = ord("?")
name2codepoint["at_sign"] = ord("@")
name2codepoint["caret"] = ord("^")
name2codepoint["pipe"] = ord("|")
name2codepoint["tilde"] = ord("~")

# These are just vexing!
del name2codepoint["and"]
del name2codepoint["or"]


def preprocess():
    if any(src in args.dependencies for src in args.changed_sources):
        sources = args.sources
    elif any(args.changed_sources):
        sources = args.changed_sources
    else:
        sources = args.sources
    csources = []
    cxxsources = []
    for source in sources:
        if source.endswith(".cpp"):
            cxxsources.append(source)
        else:
            csources.append(source)
    try:
        os.makedirs(os.path.dirname(args.output[0]))
    except OSError:
        pass

    def pp(flags):
        def run(files):
            return subprocess.check_output(args.pp + flags + files)

        return run

    try:
        cpus = multiprocessing.cpu_count()
    except NotImplementedError:
        cpus = 1
    p = multiprocessing.dummy.Pool(cpus)
    with open(args.output[0], "wb") as out_file:
        for flags, sources in (
            (args.cflags, csources),
            (args.cxxflags, cxxsources),
        ):
            batch_size = (len(sources) + cpus - 1) // cpus
            chunks = [sources[i : i + batch_size] for i in range(0, len(sources), batch_size or 1)]
            for output in p.imap(pp(flags), chunks):
                out_file.write(output)


def write_out(fname, output):
    if output:
        for m, r in [("/", "__"), ("\\", "__"), (":", "@"), ("..", "@@")]:
            fname = fname.replace(m, r)
        with open(args.output_dir + "/" + fname + ".qstr", "w") as f:
            f.write("\n".join(output) + "\n")


def qstr_unescape(qstr):
    for name in name2codepoint:
        if "__" + name + "__" in qstr:
            continue
        if "_" + name + "_" in qstr:
            qstr = qstr.replace("_" + name + "_", str(unichr(name2codepoint[name])))
    return qstr


def process_file(f):
    re_line = re.compile(r"#[line]*\s(\d+)\s\"([^\"]+)\"")
    re_qstr = re.compile(r"MP_QSTR_[_a-zA-Z0-9]+")
    re_translate = re.compile(r"translate\(\"((?:(?=(\\?))\2.)*?)\"\)")
    output = []
    last_fname = None
    lineno = 0
    for line in f:
        if line.isspace():
            continue
        # match gcc-like output (# n "file") and msvc-like output (#line n "file")
        if line.startswith(("# ", "#line")):
            m = re_line.match(line)
            assert m is not None
            lineno = int(m.group(1))
            fname = m.group(2)
            if os.path.splitext(fname)[1] not in [".c", ".cpp"]:
                continue
            if fname != last_fname:
                write_out(last_fname, output)
                output = []
                last_fname = fname
            continue
        for match in re_qstr.findall(line):
            name = match.replace("MP_QSTR_", "")
            if name not in QSTRING_BLOCK_LIST:
                output.append("Q(" + qstr_unescape(name) + ")")
        for match in re_translate.findall(line):
            output.append('TRANSLATE("' + match[0] + '")')
        lineno += 1

    if last_fname:
        write_out(last_fname, output)
    return ""


def cat_together():
    import glob
    import hashlib

    hasher = hashlib.md5()
    all_lines = []
    outf = open(args.output_dir + "/out", "wb")
    for fname in glob.glob(args.output_dir + "/*.qstr"):
        with open(fname, "rb") as f:
            lines = f.readlines()
            all_lines += lines
    all_lines.sort()
    all_lines = b"\n".join(all_lines)
    outf.write(all_lines)
    outf.close()
    hasher.update(all_lines)
    new_hash = hasher.hexdigest()
    # print(new_hash)
    old_hash = None
    try:
        with open(args.output_file + ".hash") as f:
            old_hash = f.read()
    except IOError:
        pass
    if old_hash != new_hash:
        print("QSTR updated")
        try:
            # rename below might fail if file exists
            os.remove(args.output_file)
        except:
            pass
        os.rename(args.output_dir + "/out", args.output_file)
        with open(args.output_file + ".hash", "w") as f:
            f.write(new_hash)
    else:
        print("QSTR not updated")


if __name__ == "__main__":
    if len(sys.argv) != 5:
        print("usage: %s command input_filename output_dir output_file" % sys.argv[0])
        sys.exit(2)

    class Args:
        pass

    args = Args()
    args.command = sys.argv[1]
    args.input_filename = sys.argv[2]
    args.output_dir = sys.argv[3]
    args.output_file = sys.argv[4]

    try:
        os.makedirs(args.output_dir)
    except OSError:
        pass

    if args.command == "split":
        with io.open(args.input_filename, encoding="utf-8") as infile:
            process_file(infile)

    if args.command == "cat":
        cat_together()



d = {1: 2, 3: 4}
print(d.pop(3), d)
print(d)
print(d.pop(1, 42), d)
print(d.pop(1, 42), d)
print(d.pop(1, None), d)
try:
    print(d.pop(1), "!!!")
except KeyError:
    print("Raised KeyError")
else:
    print("Did not raise KeyError!")



def gen():
    yield 1
    yield 2
    raise ValueError

def gen2():
    try:
        print((yield from gen()))
    except ValueError:
        print("caught ValueError from downstream")

g = gen2()
print(list(g))



# test construction of bytearray from different objects

print(bytearray(b'123'))
print(bytearray('1234', 'utf-8'))
print(bytearray('12345', 'utf-8', 'strict'))
print(bytearray((1, 2)))
print(bytearray([1, 2]))



# test set binary operations

sets = [set(), {1}, {1, 2}, {1, 2, 3}, {2, 3}, {2, 3, 5}, {5}, {7}]
for s in sets:
    for t in sets:
        print(sorted(s), '|', sorted(t), '=', sorted(s | t))
        print(sorted(s), '^', sorted(t), '=', sorted(s ^ t))
        print(sorted(s), '&', sorted(t), '=', sorted(s & t))
        print(sorted(s), '-', sorted(t), '=', sorted(s - t))
        u = s.copy()
        u |= t
        print(sorted(s), "|=", sorted(t), '-->', sorted(u))
        u = s.copy()
        u ^= t
        print(sorted(s), "^=", sorted(t), '-->', sorted(u))
        u = s.copy()
        u &= t
        print(sorted(s), "&=", sorted(t), "-->", sorted(u))
        u = s.copy()
        u -= t
        print(sorted(s), "-=", sorted(t), "-->", sorted(u))

        print(sorted(s), '==', sorted(t), '=', s == t)
        print(sorted(s), '!=', sorted(t), '=', s != t)
        print(sorted(s), '>', sorted(t), '=', s > t)
        print(sorted(s), '>=', sorted(t), '=', s >= t)
        print(sorted(s), '<', sorted(t), '=', s < t)
        print(sorted(s), '<=', sorted(t), '=', s <= t)

print(set('abc') == 1)

# make sure inplace operators modify the set

s1 = s2 = set('abc')
s1 |= set('ad')
print(s1 is s2, len(s1))

s1 = s2 = set('abc')
s1 ^= set('ad')
print(s1 is s2, len(s1))

s1 = s2 = set('abc')
s1 &= set('ad')
print(s1 is s2, len(s1))

s1 = s2 = set('abc')
s1 -= set('ad')
print(s1 is s2, len(s1))

# RHS must be a set
try:
    print(set('12') >= '1')
except TypeError:
    print('TypeError')

# RHS must be a set
try:
    print(set('12') <= '123')
except TypeError:
    print('TypeError')

# unsupported operator
try:
    set('abc') * set('abc')
except TypeError:
    print('TypeError')

# unsupported operator with RHS not a set
try:
    set('abc') * 2
except TypeError:
    print('TypeError')



# Test that native code loaded from a .mpy file is retained after a GC.

try:
    import gc, sys, uio, uos

    sys.implementation.mpy
    uio.IOBase
    uos.mount
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class UserFile(uio.IOBase):
    def __init__(self, data):
        self.data = memoryview(data)
        self.pos = 0

    def readinto(self, buf):
        n = min(len(buf), len(self.data) - self.pos)
        buf[:n] = self.data[self.pos : self.pos + n]
        self.pos += n
        return n

    def ioctl(self, req, arg):
        return 0


class UserFS:
    def __init__(self, files):
        self.files = files

    def mount(self, readonly, mksfs):
        pass

    def umount(self):
        pass

    def stat(self, path):
        if path in self.files:
            return (32768, 0, 0, 0, 0, 0, 0, 0, 0, 0)
        raise OSError

    def open(self, path, mode):
        return UserFile(self.files[path])


# Pre-compiled examples/natmod/features0 example for various architectures, keyed
# by the required value of sys.implementation.mpy.
features0_file_contents = {
    # -march=x64 -mcache-lookup-bc
    0xB05: b'C\x05\x0b\x1f \x84b\xe9/\x00\x00\x00SH\x8b\x1ds\x00\x00\x00\xbe\x02\x00\x00\x00\xffS\x18\xbf\x01\x00\x00\x00H\x85\xc0u\x0cH\x8bC \xbe\x02\x00\x00\x00[\xff\xe0H\x0f\xaf\xf8H\xff\xc8\xeb\xe6ATUSH\x8b\x1dA\x00\x00\x00H\x8b\x7f\x08H\x8bk(\xff\xd5H\x8d5 \x00\x00\x00I\x89\xc4H\x8b\x05.\x00\x00\x00\x0f\xb78\xffShL\x89\xe7\xff\xd5H\x8b\x03[]A\\\xc3\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x90\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x84@\x12factorial \x00\x00\r \x01"\xa1\x1c\x01\x1e\xff',
    # -march=armv7m
    0x1605: b"C\x05\x16\x1f \x84\x12\x1a\xe0\x00\x00\x13\xb5\nK\nJ{D\x9cX\x02!\xe3h\x98G\x03F\x01 3\xb9\x02!#i\x01\x93\x02\xb0\xbd\xe8\x10@\x18GXC\x01;\xf4\xe7\x00\xbfj\x00\x00\x00\x00\x00\x00\x00\xf8\xb5\tN\tK~D\xf4X@hgi\xb8G\x05F\x07K\x07I\xf2XyD\x10\x88ck\x98G(F\xb8G h\xf8\xbd6\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x1c\x00\x00\x00\x00\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x00\x00\x00\x00\x01\x84\x00\x12factorial \x00\x00\r<\x01>\xa18\x01:\xff",
}

# Populate other armv7m-derived archs based on armv7m.
for arch in (0x1A05, 0x1E05, 0x2205):
    features0_file_contents[arch] = features0_file_contents[0x1605]

if sys.implementation.mpy not in features0_file_contents:
    print("SKIP")
    raise SystemExit

# These are the test .mpy files.
user_files = {"/features0.mpy": features0_file_contents[sys.implementation.mpy]}

# Create and mount a user filesystem.
uos.mount(UserFS(user_files), "/userfs")
sys.path.append("/userfs")

# Import the native function.
gc.collect()
from features0 import factorial

# Free the module that contained the function.
del sys.modules["features0"]

# Run a GC cycle which should reclaim the module but not the function.
gc.collect()

# Allocate lots of fragmented memory to overwrite anything that was just freed by the GC.
for i in range(1000):
    []

# Run the native function, it should not have been freed or overwritten.
print(factorial(10))

# Unmount and undo path addition.
uos.umount("/userfs")
sys.path.pop()



# test named char classes

try:
    import ure as re
except ImportError:
    try:
        import re
    except ImportError:
        print("SKIP")
        raise SystemExit


def print_groups(match):
    print("----")
    try:
        i = 0
        while True:
            print(m.group(i))
            i += 1
    except IndexError:
        pass


m = re.match(r"\w+", "1234hello567 abc")
print_groups(m)

m = re.match(r"(\w+)\s+(\w+)", "ABC \t1234hello567 abc")
print_groups(m)

m = re.match(r"(\S+)\s+(\D+)", "ABC \thello abc567 abc")
print_groups(m)

m = re.match(r"(([0-9]*)([a-z]*)\d*)", "1234hello567")
print_groups(m)



# test arithmetic operators with uint type


@micropython.viper
def add(x: uint, y: uint):
    return x + y, y + x


print("add")
print(*add(1, 2))
print(*(x & 0xFFFFFFFF for x in add(-1, -2)))


@micropython.viper
def sub(x: uint, y: uint):
    return x - y, y - x


print("sub")
print(*(x & 0xFFFFFFFF for x in sub(1, 2)))
print(*(x & 0xFFFFFFFF for x in sub(-1, -2)))


@micropython.viper
def mul(x: uint, y: uint):
    return x * y, y * x


print("mul")
print(*mul(2, 3))
print(*(x & 0xFFFFFFFF for x in mul(2, -3)))
print(*mul(-2, -3))



# test that the GC can trace nested objects

try:
    import gc
except ImportError:
    print("SKIP")
    raise SystemExit

# test a big shallow object pointing to many unique objects
lst = [[i] for i in range(200)]
gc.collect()
print(lst)

# test a deep object
lst = [
    [[[[(i, j, k, l)] for i in range(3)] for j in range(3)] for k in range(3)] for l in range(3)
]
gc.collect()
print(lst)



# test 'return' within the finally block, with nested finally's
# only inactive finally's should be executed, and only once

# basic nested finally's, the print should only be executed once
def f():
    try:
        raise TypeError
    finally:
        print(1)
        try:
            raise ValueError
        finally:
            return 42
print(f())

# similar to above but more nesting
def f():
    try:
        raise ValueError
    finally:
        print(1)
        try:
            raise TypeError
        finally:
            print(2)
            try:
                pass
            finally:
                return 42
print(f())

# similar to above but even more nesting
def f():
    try:
        raise ValueError
    finally:
        print(1)
        try:
            raise TypeError
        finally:
            print(2)
            try:
                raise Exception
            finally:
                print(3)
                return 42
print(f())

# upon return some try's are active, some finally's are active, some inactive
def f():
    try:
        try:
            pass
        finally:
            print(2)
            return 42
    finally:
        print(1)
print(f())

# same as above but raise instead of pass
def f():
    try:
        try:
            raise ValueError
        finally:
            print(2)
            return 42
    finally:
        print(1)
print(f())

# upon return exception stack holds: active finally, inactive finally, active finally
def f():
    try:
        raise Exception
    finally:
        print(1)
        try:
            try:
                pass
            finally:
                print(3)
                return 42
        finally:
            print(2)
print(f())

# same as above but raise instead of pass in innermost try block
def f():
    try:
        raise Exception
    finally:
        print(1)
        try:
            try:
                raise Exception
            finally:
                print(3)
                return 42
        finally:
            print(2)
print(f())



f = open("io/data/file1", "rb")
print(f.seek(6))
print(f.read(5))
print(f.tell())

print(f.seek(0, 1))
print(f.read(4))
print(f.tell())

print(f.seek(-6, 2))
print(f.read(20))
print(f.tell())

print(f.seek(0, 0))
print(f.read(5))
print(f.tell())

f.close()

# test text mode
f = open("io/data/file1", "rt")
print(f.seek(6))
print(f.read(5))
print(f.tell())
f.close()

# seek closed file
f = open("io/data/file1", "r")
f.close()
try:
    f.seek(1)
except (OSError, ValueError):
    # CPy raises ValueError, uPy raises OSError
    print("OSError or ValueError")



import bench


class Foo:
    def __init__(self):
        self._num = 20000000

    def num(self):
        return self._num


def test(num):
    o = Foo()
    i = 0
    while i < o.num():
        i += 1


bench.run(test)



# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# SPDX-FileCopyrightText: Copyright (c) 2017 Scott Shawcroft for Adafruit Industries
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This helper generates the pinout tables in ../README.rst.

import os
import os.path

pins = [
    "PA00",
    "PA01",
    "PA02",
    "PA03",
    "PB08",
    "PB09",
    "PA04",
    "PA05",
    "PA06",
    "PA07",
    "PA08",
    "PA09",
    "PA10",
    "PA11",
    "PB10",
    "PB11",
    "PA12",
    "PA13",
    "PA14",
    "PA15",
    "PA16",
    "PA17",
    "PA18",
    "PA19",
    "PA20",
    "PA21",
    "PA22",
    "PA23",
    "PA24",
    "PA25",
    "PB22",
    "PB23",
    "PA27",
    "PA28",
    "PA29",
    "PA30",
    "PA31",
    "PB02",
    "PB03",
]

# Dictionary keys: [board][pin] = list of pin names
mapping = {}

QSTR = "    { MP_OBJ_NEW_QSTR(MP_QSTR_"

for board in os.listdir("boards"):
    if not os.path.isdir("boards/" + board):
        continue
    mapping[board] = {}
    with open("boards/" + board + "/pins.c", "r") as f:
        for line in f:
            if line.startswith(QSTR):
                board_name, _, pin = line.split(")")
                board_name = board_name[len(QSTR) :]
                pin = pin[-8:-4]
                if pin not in mapping[board]:
                    mapping[board][pin] = []
                mapping[board][pin].append(board_name)

column_width = {}
for board in mapping:
    column_width[board] = len(board)
    for pin in mapping[board]:
        l = len(" / ".join("``" + x + "``" for x in mapping[board][pin]))
        column_width[board] = max(l, column_width[board])

first_column_width = len("`microcontroller.pin`")
print("=" * first_column_width, end="")
total_board_width = -2
for board in column_width:
    column = "  " + "=" * column_width[board]
    total_board_width += len(column)
    print(column, end="")

print()
print("`microcontroller.pin`  `board`")
print("-" * first_column_width + "  " + "-" * total_board_width)

print("Datasheet".ljust(first_column_width), end="")
for board in column_width:
    print("  " + board.ljust(column_width[board]), end="")
print()

print("=" * first_column_width, end="")
for board in column_width:
    column = "  " + "=" * column_width[board]
    print(column, end="")
print()

for pin in pins:
    print(pin.ljust(first_column_width), end="")
    for board in column_width:
        if pin in mapping[board]:
            names = " / ".join("``" + x + "``" for x in mapping[board][pin])
            print("  " + names.ljust(column_width[board]), end="")
        else:
            print(" " * (column_width[board] + 2), end="")
    print()

print("=" * first_column_width, end="")
for board in column_width:
    column = "  " + "=" * column_width[board]
    print(column, end="")
print()

print()
print()
# Generate pin capabilities too.

ALL_BUT_USB = list(pins)
ALL_BUT_USB.remove("PA24")
ALL_BUT_USB.remove("PA25")

# dictionary is [module][class] = [pins]
capabilities = {
    "analogio": {
        "AnalogIn": [
            "PA02",
            "PA03",
            "PB08",
            "PB09",
            "PA04",
            "PA05",
            "PA06",
            "PA07",
            "PA08",
            "PA09",
            "PA10",
            "PA11",
            "PB02",
            "PB03",
        ],
        "AnalogOut": ["PA02"],
    },
    "audioio": {"AudioOut": ["PA02"]},
    "bitbangio": {"I2C": ALL_BUT_USB, "OneWire": ALL_BUT_USB, "SPI": ALL_BUT_USB},
    "busio": {
        "I2C - SDA": ["PA00", "PB08", "PA08", "PA12", "PA16", "PA22", "PB02"],  # SERCOM pad 0
        "I2C - SCL": ["PA01", "PB09", "PA09", "PA13", "PA17", "PA23", "PB03"],  # SERCOM pad 1
        "OneWire": ALL_BUT_USB,
        "SPI - MISO": [
            "PA00",
            "PA01",
            "PB08",
            "PB09",
            "PA04",
            "PA05",
            "PA06",
            "PA07",
            "PA08",
            "PA09",
            "PA10",
            "PA11",
            "PB10",
            "PB11",
            "PA12",
            "PA13",
            "PA14",
            "PA15",
            "PA16",
            "PA17",
            "PA18",
            "PA19",
            "PA20",
            "PA21",
            "PA22",
            "PA23",
            "PB22",
            "PB23",
            "PA30",
            "PA31",
            "PB02",
            "PB03",
        ],  # any SERCOM pad
        "SPI - MOSI": [
            "PA00",
            "PB08",
            "PA04",
            "PA06",
            "PA08",
            "PA10",
            "PA11",
            "PB10",
            "PB11",
            "PA14",
            "PA15",
            "PA16",
            "PA18",
            "PA19",
            "PA20",
            "PA21",
            "PA22",
            "PB22",
            "PB23",
            "PA30",
            "PA31",
            "PB02",
        ],  # any pad but 1
        "SPI - SCK": [
            "PA01",
            "PB09",
            "PA05",
            "PA07",
            "PA09",
            "PA11",
            "PB11",
            "PA13",
            "PA15",
            "PA17",
            "PA19",
            "PA21",
            "PA23",
            "PB23",
            "PA31",
            "PB03",
        ],  # 1 or 3
        "UART - RX": [
            "PA00",
            "PA01",
            "PB08",
            "PB09",
            "PA04",
            "PA05",
            "PA06",
            "PA07",
            "PA08",
            "PA09",
            "PA10",
            "PA11",
            "PB10",
            "PB11",
            "PA12",
            "PA13",
            "PA14",
            "PA15",
            "PA16",
            "PA17",
            "PA18",
            "PA19",
            "PA20",
            "PA21",
            "PA22",
            "PA23",
            "PB22",
            "PB23",
            "PA30",
            "PA31",
            "PB02",
            "PB03",
        ],  # any pad
        "UART - TX": [
            "PA00",
            "PB08",
            "PA04",
            "PA06",
            "PA08",
            "PA10",
            "PB10",
            "PA12",
            "PA14",
            "PA16",
            "PA18",
            "PA20",
            "PA22",
            "PB22",
            "PA30",
            "PB02",
        ],  # pad 0 or 2
    },
    "digitalio": {"DigitalInOut": ALL_BUT_USB},
    "pulseio": {
        "PulseIn": ALL_BUT_USB,
        "PWMOut": [
            "PA01",
            "PB09",
            "PA04",
            "PA05",
            "PA06",
            "PA07",
            "PA08",
            "PA09",
            "PA10",
            "PA11",
            "PB10",
            "PB11",
            "PA12",
            "PA13",
            "PA14",
            "PA15",
            "PA16",
            "PA17",
            "PA18",
            "PA19",
            "PA20",
            "PA21",
            "PA22",
            "PA23",
            "PA30",
            "PA31",
        ],
    },
    "ps2io": {"Ps2": ALL_BUT_USB},
    "touchio": {
        "TouchIn": ["PA02", "PA03", "PB08", "PB09", "PA04", "PA05", "PA06", "PA07", "PB02", "PB03"]
    },
}

column_width = {}
for module in capabilities:
    for c in capabilities[module]:
        column_width[module + c] = max(len("**Yes**"), len(c))

module_width = {}
for module in capabilities:
    module_width[module] = 0
    for c in capabilities[module]:
        module_width[module] += column_width[module + c] + 2
    module_width[module] -= 2

    if module_width[module] < (len(module) + 2):
        column_width[module + c] += len(module) + 2 - module_width[module]
        module_width[module] = len(module) + 2

first_column_width = len("`microcontroller.pin`")
print("=" * first_column_width, end="")
for module in capabilities:
    for c in capabilities[module]:
        print("  " + "=" * column_width[module + c], end="")
print()

print("`microcontroller.pin`", end="")
for module in capabilities:
    print("  " + ("`" + module + "`").ljust(module_width[module]), end="")
print()

print("-" * first_column_width, end="")
for module in capabilities:
    print("  " + "-" * module_width[module], end="")
print()

print("Datasheet".ljust(first_column_width), end="")
for module in capabilities:
    for c in capabilities[module]:
        print("  " + c.ljust(column_width[module + c]), end="")
print()

print("=" * first_column_width, end="")
for module in capabilities:
    for c in capabilities[module]:
        print("  " + "=" * column_width[module + c], end="")
print()

for pin in pins:
    print(pin.ljust(first_column_width), end="")
    for module in capabilities:
        for c in capabilities[module]:
            if pin in capabilities[module][c]:
                print("  " + "**Yes**".ljust(column_width[module + c]), end="")
            else:
                print(" " * (column_width[module + c] + 2), end="")
    print()

print("=" * first_column_width, end="")
for module in capabilities:
    for c in capabilities[module]:
        print("  " + "=" * column_width[module + c], end="")
print()



# test bytes + other
try:
    import uarray as array
except ImportError:
    try:
        import array
    except ImportError:
        print("SKIP")
        raise SystemExit

print(b"123" + array.array('i', [1]))



import skip_if
skip_if.no_cpython_compat()

try:
    from collections import namedtuple
except ImportError:
    skip_if.skip()

# Try single string
T3 = namedtuple("TupComma", "foo bar")
t = T3(1, 2)
print(t.foo, t.bar)

# Try single string with comma field separator
# Not implemented so far
#T2 = namedtuple("TupComma", "foo,bar")
#t = T2(1, 2)



def Fun():
    pass

class A:
    def __init__(self):
        pass
    def Fun(self):
        pass

try:
    print(Fun.__name__)
    print(A.__init__.__name__)
    print(A.Fun.__name__)
    print(A().Fun.__name__)
except AttributeError:
    print('SKIP')
    raise SystemExit

# __name__ of a bound native method is not implemented in uPy
# the test here is to make sure it doesn't crash
try:
    str((1).to_bytes.__name__)
except AttributeError:
    pass

# name of a function that has closed over variables
def outer():
    x = 1
    def inner():
        return x
    return inner
print(outer.__name__)



# test capability for threads to access a shared immutable data structure
#
# SPDX-FileCopyrightText: Copyright (c) 2016 Damien P. George on behalf of Pycom Ltd
#
# SPDX-License-Identifier: MIT

import _thread


def foo(i):
    pass


def thread_entry(n, tup):
    for i in tup:
        foo(i)
    with lock:
        global n_finished
        n_finished += 1


lock = _thread.allocate_lock()
n_thread = 2
n_finished = 0

# the shared data structure
tup = (1, 2, 3, 4)

# spawn threads
for i in range(n_thread):
    _thread.start_new_thread(thread_entry, (100, tup))

# busy wait for threads to finish
while n_finished < n_thread:
    pass
print(tup)



# Test that returning of NotImplemented from binary op methods leads to
# TypeError.
try:
    NotImplemented
except NameError:
    print("SKIP")
    raise SystemExit

class C:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "C({})".format(self.value)

    def __add__(self, rhs):
        print(self, '+', rhs)
        return NotImplemented

    def __sub__(self, rhs):
        print(self, '-', rhs)
        return NotImplemented

    def __lt__(self, rhs):
        print(self, '<', rhs)
        return NotImplemented

    def __neg__(self):
        print('-', self)
        return NotImplemented

c = C(0)

try:
    c + 1
except TypeError:
    print("TypeError")

try:
    c - 2
except TypeError:
    print("TypeError")

try:
    c < 1
except TypeError:
    print("TypeError")

# NotImplemented isn't handled specially in unary methods
print(-c)

# Test that NotImplemented can be hashed
print(type(hash(NotImplemented)))



# test floor-division and modulo operators


@micropython.viper
def div(x: int, y: int) -> int:
    return x // y


@micropython.viper
def mod(x: int, y: int) -> int:
    return x % y


def dm(x, y):
    print(div(x, y), mod(x, y))


for x in (-6, 6):
    for y in range(-7, 8):
        if y == 0:
            continue
        dm(x, y)



#! /usr/bin/env python3

# Copyright (C) 2018 Sony Semiconductor Solutions Corp.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 3. Neither the name NuttX nor the names of its contributors may be
#    used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import time
import sys
import os
import struct
import glob
import fnmatch
import errno
import telnetlib
import argparse
import shutil
import subprocess
import re
import xmodem

import_serial_module = True

# When SDK release, plase set SDK_RELEASE as True.
SDK_RELEASE = False

if SDK_RELEASE:
    PRINT_RAW_COMMAND = False
    REBOOT_AT_END = True
else:
    PRINT_RAW_COMMAND = True
    REBOOT_AT_END = True

try:
    import serial
except:
    import_serial_module = False

# supported environment various
# CXD56_PORT
# CXD56_TELNETSRV_PORT
# CXD56_TELNETSRV_IP

PROTOCOL_SERIAL = 0
PROTOCOL_TELNET = 1

MAX_DOT_COUNT = 70

# configure parameters and default value
class ConfigArgs:
    PROTOCOL_TYPE = None
    SERIAL_PORT = "COM1"
    SERVER_PORT = 4569
    SERVER_IP = "localhost"
    EOL = bytes([10])
    WAIT_RESET = True
    AUTO_RESET = False
    DTR_RESET = False
    XMODEM_BAUD = 0
    NO_SET_BOOTABLE = False
    PACKAGE_NAME = []
    FILE_NAME = []
    ERASE_NAME = []
    PKGSYS_NAME = []
    PKGAPP_NAME = []
    PKGUPD_NAME = []


ROM_MSG = [b"Welcome to nash"]
XMDM_MSG = "Waiting for XMODEM (CRC or 1K) transfer. Ctrl-X to cancel."


class ConfigArgsLoader:
    def __init__(self):
        self.parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
        self.parser.add_argument(
            "package_name", help="the name of the package to install", nargs="*"
        )
        self.parser.add_argument(
            "-f", "--file", dest="file_name", help="save file", action="append"
        )
        self.parser.add_argument(
            "-e", "--erase", dest="erase_name", help="erase file", action="append"
        )

        self.parser.add_argument(
            "-S",
            "--sys",
            dest="pkgsys_name",
            help="the name of the system package to install",
            action="append",
        )
        self.parser.add_argument(
            "-A",
            "--app",
            dest="pkgapp_name",
            help="the name of the application package to install",
            action="append",
        )
        self.parser.add_argument(
            "-U",
            "--upd",
            dest="pkgupd_name",
            help="the name of the updater package to install",
            action="append",
        )

        self.parser.add_argument(
            "-a",
            "--auto-reset",
            dest="auto_reset",
            action="store_true",
            default=None,
            help="try to auto reset develop board if possible",
        )
        self.parser.add_argument(
            "-d",
            "--dtr-reset",
            dest="dtr_reset",
            action="store_true",
            default=None,
            help="try to auto reset develop board if possible",
        )
        self.parser.add_argument(
            "-n",
            "--no-set-bootable",
            dest="no_set_bootable",
            action="store_true",
            default=None,
            help="not to set bootable",
        )

        group = self.parser.add_argument_group()
        group.add_argument(
            "-i",
            "--server-ip",
            dest="server_ip",
            help="the ip address connected to the telnet server",
        )
        group.add_argument(
            "-p",
            "--server-port",
            dest="server_port",
            type=int,
            help="the port connected to the telnet server",
        )

        group = self.parser.add_argument_group()
        group.add_argument("-c", "--serial-port", dest="serial_port", help="the serial port")
        group.add_argument(
            "-b", "--xmodem-baudrate", dest="xmodem_baud", help="Use the faster baudrate in xmodem"
        )

        mutually_group = self.parser.add_mutually_exclusive_group()
        mutually_group.add_argument(
            "-t",
            "--telnet-protocol",
            dest="telnet_protocol",
            action="store_true",
            default=None,
            help="use the telnet protocol for binary transmission",
        )
        mutually_group.add_argument(
            "-s",
            "--serial-protocol",
            dest="serial_protocol",
            action="store_true",
            default=None,
            help="use the serial port for binary transmission, default options",
        )

        mutually_group2 = self.parser.add_mutually_exclusive_group()
        mutually_group2.add_argument(
            "-F",
            "--force-wait-reset",
            dest="wait_reset",
            action="store_true",
            default=None,
            help="force wait for pressing RESET button",
        )
        mutually_group2.add_argument(
            "-N",
            "--no-wait-reset",
            dest="wait_reset",
            action="store_false",
            default=None,
            help="if possible, skip to wait for pressing RESET button",
        )

    def update_config(self):
        args = self.parser.parse_args()

        ConfigArgs.PACKAGE_NAME = args.package_name
        ConfigArgs.FILE_NAME = args.file_name
        ConfigArgs.ERASE_NAME = args.erase_name
        ConfigArgs.PKGSYS_NAME = args.pkgsys_name
        ConfigArgs.PKGAPP_NAME = args.pkgapp_name
        ConfigArgs.PKGUPD_NAME = args.pkgupd_name

        # Get serial port or telnet server ip etc
        if args.serial_protocol == True:
            ConfigArgs.PROTOCOL_TYPE = PROTOCOL_SERIAL
        elif args.telnet_protocol == True:
            ConfigArgs.PROTOCOL_TYPE = PROTOCOL_TELNET

        if ConfigArgs.PROTOCOL_TYPE == None:
            proto = os.environ.get("CXD56_PROTOCOL")
            if proto is not None:
                if "s" in proto:
                    ConfigArgs.PROTOCOL_TYPE = PROTOCOL_SERIAL
                elif "t" in proto:
                    ConfigArgs.PROTOCOL_TYPE = PROTOCOL_TELNET

        if ConfigArgs.PROTOCOL_TYPE == None:
            ConfigArgs.PROTOCOL_TYPE = PROTOCOL_SERIAL

        if ConfigArgs.PROTOCOL_TYPE == PROTOCOL_SERIAL:
            if args.serial_port is not None:
                ConfigArgs.SERIAL_PORT = args.serial_port
            else:
                # Get serial port from the environment
                port = os.environ.get("CXD56_PORT")
                if port is not None:
                    ConfigArgs.SERIAL_PORT = port
                else:
                    print("CXD56_PORT is not set, Use " + ConfigArgs.SERIAL_PORT + ".")
        else:
            ConfigArgs.PROTOCOL_TYPE = PROTOCOL_TELNET
            if args.server_port is not None:
                ConfigArgs.SERVER_PORT = args.server_port
            else:
                port = os.environ.get("CXD56_TELNETSRV_PORT")
                if port is not None:
                    ConfigArgs.SERVER_PORT = port
                else:
                    print(
                        "CXD56_TELNETSRV_PORT is not set, Use " + str(ConfigArgs.SERVER_PORT) + "."
                    )
            if args.server_ip is not None:
                ConfigArgs.SERVER_IP = args.server_ip
            else:
                ip = os.environ.get("CXD56_TELNETSRV_IP")
                if ip is not None:
                    ConfigArgs.SERVER_IP = ip
                else:
                    print("CXD56_TELNETSRV_IP is not set, Use " + ConfigArgs.SERVER_IP + ".")

        if args.xmodem_baud is not None:
            ConfigArgs.XMODEM_BAUD = args.xmodem_baud

        if args.auto_reset is not None:
            ConfigArgs.AUTO_RESET = args.auto_reset

        if args.dtr_reset is not None:
            ConfigArgs.DTR_RESET = args.dtr_reset

        if args.no_set_bootable is not None:
            ConfigArgs.NO_SET_BOOTABLE = args.no_set_bootable

        if args.wait_reset is not None:
            ConfigArgs.WAIT_RESET = args.wait_reset


class TelnetDev:
    def __init__(self):
        srv_ipaddr = ConfigArgs.SERVER_IP
        srv_port = ConfigArgs.SERVER_PORT
        self.recvbuf = b""
        try:
            self.telnet = telnetlib.Telnet(host=srv_ipaddr, port=srv_port, timeout=10)
            # There is a ack to be sent after connecting to the telnet server.
            self.telnet.write(b"\xff")
        except Exception as e:
            print("Cannot connect to the server %s:%d" % (srv_ipaddr, srv_port))
            sys.exit(e.args[0])

    def readline(self, size=None):
        res = b""
        ch = b""
        while ch != ConfigArgs.EOL:
            ch = self.getc_raw(1, timeout=0.1)
            if ch == b"":
                return res
            res += ch
        return res

    def getc_raw(self, size, timeout=1):
        res = b""
        tm = time.monotonic()
        while size > 0:
            while self.recvbuf == b"":
                self.recvbuf = self.telnet.read_eager()
                if self.recvbuf == b"":
                    if (time.monotonic() - tm) > timeout:
                        return res
                    time.sleep(0.1)
            res += self.recvbuf[0:1]
            self.recvbuf = self.recvbuf[1:]
            size -= 1
        return res

    def write(self, buffer):
        self.telnet.write(buffer)

    def discard_inputs(self, timeout=1.0):
        while True:
            ch = self.getc_raw(1, timeout=timeout)
            if ch == b"":
                break

    def getc(self, size, timeout=1):
        c = self.getc_raw(size, timeout)
        return c

    def putc(self, buffer, timeout=1):
        self.telnet.write(buffer)
        self.show_progress(len(buffer))

    def reboot(self):
        # no-op
        pass

    def set_file_size(self, filesize):
        self.bytes_transfered = 0
        self.filesize = filesize
        self.count = 0

    def show_progress(self, sendsize):
        if PRINT_RAW_COMMAND:
            if self.count < MAX_DOT_COUNT:
                self.bytes_transfered = self.bytes_transfered + sendsize
                cur_count = int(self.bytes_transfered * MAX_DOT_COUNT / self.filesize)
                if MAX_DOT_COUNT < cur_count:
                    cur_count = MAX_DOT_COUNT
                for idx in range(cur_count - self.count):
                    print("#", end="")
                    sys.stdout.flush()
                self.count = cur_count
                if self.count == MAX_DOT_COUNT:
                    print("\n")


class SerialDev:
    def __init__(self):
        if import_serial_module is False:
            print("Cannot import serial module, maybe it's not install yet.")
            print("\n", end="")
            print("Please install python-setuptool by Cygwin installer.")
            print("After that use easy_intall command to install serial module")
            print("    $ cd tool/")
            print("    $ python3 -m easy_install pyserial-2.7.tar.gz")
            quit()
        else:
            port = ConfigArgs.SERIAL_PORT
            try:
                self.serial = serial.Serial(
                    port,
                    baudrate=115200,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                    bytesize=serial.EIGHTBITS,
                    timeout=0.1,
                )
            except Exception as e:
                print("Cannot open port : " + port)
                sys.exit(e.args[0])

    def readline(self, size=None):
        return self.serial.readline(size)

    def write(self, buffer):
        self.serial.write(buffer)
        self.serial.flush()

    def discard_inputs(self, timeout=1.0):
        time.sleep(timeout)
        self.serial.flushInput()

    def getc(self, size, timeout=1):
        self.serial.timeout = timeout
        c = self.serial.read(size)
        self.serial.timeout = 0.1
        return c

    def putc(self, buffer, timeout=1):
        self.serial.timeout = timeout
        self.serial.write(buffer)
        self.serial.flush()
        self.serial.timeout = 0.1
        self.show_progress(len(buffer))

    # Note: windows platform dependent code
    def putc_win(self, buffer, timeout=1):
        self.serial.write(buffer)
        self.show_progress(len(buffer))
        while True:
            if self.serial.out_waiting == 0:
                break

    def setBaudrate(self, baudrate):
        # 		self.serial.setBaudrate(baudrate)
        self.serial.baudrate = baudrate

    def reboot(self):
        # Target Reset by DTR
        self.serial.setDTR(False)
        self.serial.setDTR(True)
        self.serial.setDTR(False)

    def set_file_size(self, filesize):
        self.bytes_transfered = 0
        self.filesize = filesize
        self.count = 0

    def show_progress(self, sendsize):
        if PRINT_RAW_COMMAND:
            if self.count < MAX_DOT_COUNT:
                self.bytes_transfered = self.bytes_transfered + sendsize
                cur_count = int(self.bytes_transfered * MAX_DOT_COUNT / self.filesize)
                if MAX_DOT_COUNT < cur_count:
                    cur_count = MAX_DOT_COUNT
                for idx in range(cur_count - self.count):
                    print("#", end="")
                    sys.stdout.flush()
                self.count = cur_count
                if self.count == MAX_DOT_COUNT:
                    print("\n")


class FlashWriter:
    def __init__(self, protocol_sel=PROTOCOL_SERIAL):
        if protocol_sel == PROTOCOL_TELNET:
            self.serial = TelnetDev()
        else:
            self.serial = SerialDev()

    def cancel_autoboot(self):
        boot_msg = ""
        self.serial.reboot()  # Target reboot before send 'r'
        while boot_msg == "":
            rx = self.serial.readline().strip()
            self.serial.write(b"r")  # Send "r" key to avoid auto boot
            for msg in ROM_MSG:
                if msg in rx:
                    boot_msg = msg
                    break
        while True:
            rx = self.serial.readline().decode(errors="replace").strip()
            if "updater" in rx:
                # Workaround : Sometime first character is dropped.
                # Send line feed as air shot before actual command.
                self.serial.write(b"\n")  # Send line feed
                self.serial.discard_inputs()  # Clear input buffer to sync
                return boot_msg.decode(errors="ignore")

    def recv(self):
        rx = self.serial.readline()
        if PRINT_RAW_COMMAND:
            serial_line = rx.decode(errors="replace")
            if serial_line.strip() != "" and not serial_line.startswith(XMDM_MSG):
                print(serial_line, end="")
        return rx

    def wait(self, string):
        while True:
            rx = self.recv()
            if string.encode() in rx:
                time.sleep(0.1)
                break

    def wait_for_prompt(self):
        prompt_pat = re.compile(b"updater")
        while True:
            rx = self.recv()
            if prompt_pat.search(rx):
                time.sleep(0.1)
                break

    def send(self, string):
        self.serial.write(str(string).encode() + b"\n")
        rx = self.serial.readline()
        if PRINT_RAW_COMMAND:
            print(rx.decode(errors="replace"), end="")

    def read_output(self, prompt_text):
        output = []
        while True:
            rx = self.serial.readline()
            if prompt_text.encode() in rx:
                time.sleep(0.1)
                break
            if rx != "":
                output.append(rx.decode(errors="ignore").rstrip())
        return output

    def install_files(self, files, command):
        if ConfigArgs.XMODEM_BAUD:
            command += " -b " + ConfigArgs.XMODEM_BAUD
        if os.name == "nt":
            modem = xmodem.XMODEM(self.serial.getc, self.serial.putc_win, "xmodem1k")
        else:
            modem = xmodem.XMODEM(self.serial.getc, self.serial.putc, "xmodem1k")
        for file in files:
            with open(file, "rb") as bin:
                self.send(command)
                print("Install " + file)
                self.wait(XMDM_MSG)
                print(
                    "|0%"
                    + "-" * (int(MAX_DOT_COUNT / 2) - 6)
                    + "50%"
                    + "-" * (MAX_DOT_COUNT - int(MAX_DOT_COUNT / 2) - 5)
                    + "100%|"
                )
                if ConfigArgs.XMODEM_BAUD:
                    self.serial.setBaudrate(ConfigArgs.XMODEM_BAUD)
                    self.serial.discard_inputs()  # Clear input buffer to sync
                self.serial.set_file_size(os.path.getsize(file))
                modem.send(bin)
                if ConfigArgs.XMODEM_BAUD:
                    self.serial.setBaudrate(115200)
            self.wait_for_prompt()

    def save_files(self, files):
        if ConfigArgs.XMODEM_BAUD:
            command = "save_file -b " + ConfigArgs.XMODEM_BAUD + " -x "
        else:
            command = "save_file -x "
        if os.name == "nt":
            modem = xmodem.XMODEM(self.serial.getc, self.serial.putc_win, "xmodem1k")
        else:
            modem = xmodem.XMODEM(self.serial.getc, self.serial.putc, "xmodem1k")
        for file in files:
            with open(file, "rb") as bin:
                self.send(command + os.path.basename(file))
                print("Save " + file)
                self.wait(XMDM_MSG)
                if ConfigArgs.XMODEM_BAUD:
                    self.serial.setBaudrate(ConfigArgs.XMODEM_BAUD)
                    self.serial.discard_inputs()  # Clear input buffer to sync
                self.serial.set_file_size(os.path.getsize(file))
                modem.send(bin)
                if ConfigArgs.XMODEM_BAUD:
                    self.serial.setBaudrate(115200)
                self.wait_for_prompt()
                self.send("chmod d+rw " + os.path.basename(file))
                self.wait_for_prompt()

    def delete_files(self, files):
        for file in files:
            self.delete_binary(file)

    def delete_binary(self, bin_name):
        self.send("rm " + bin_name)
        self.wait_for_prompt()


def main():
    try:
        config_loader = ConfigArgsLoader()
        config_loader.update_config()
    except:
        return errno.EINVAL

    # Wait to reset the board
    writer = FlashWriter(ConfigArgs.PROTOCOL_TYPE)

    do_wait_reset = True
    if ConfigArgs.AUTO_RESET:
        if subprocess.call("cd " + sys.path[0] + "; ./reset_board.sh", shell=True) == 0:
            print("auto reset board sucess!!")
            do_wait_reset = False
            bootrom_msg = writer.cancel_autoboot()

    if ConfigArgs.DTR_RESET:
        do_wait_reset = False
        bootrom_msg = writer.cancel_autoboot()

    if ConfigArgs.WAIT_RESET == False and do_wait_reset == True:
        rx = writer.recv()
        time.sleep(1)
        for i in range(3):
            writer.send("")
            rx = writer.recv()
            if "updater".encode() in rx:
                # No need to wait for reset
                do_wait_reset = False
                break
            time.sleep(1)

    if do_wait_reset:
        # Wait to reset the board
        print("Please press RESET button on target board")
        sys.stdout.flush()
        bootrom_msg = writer.cancel_autoboot()

    # Remove files
    if ConfigArgs.ERASE_NAME:
        print(">>> Remove exisiting files ...")
        writer.delete_files(ConfigArgs.ERASE_NAME)

    # Install files
    if (
        ConfigArgs.PACKAGE_NAME
        or ConfigArgs.PKGSYS_NAME
        or ConfigArgs.PKGAPP_NAME
        or ConfigArgs.PKGUPD_NAME
    ):
        print(">>> Install files ...")
    if ConfigArgs.PACKAGE_NAME:
        writer.install_files(ConfigArgs.PACKAGE_NAME, "install")
    if ConfigArgs.PKGSYS_NAME:
        writer.install_files(ConfigArgs.PKGSYS_NAME, "install")
    if ConfigArgs.PKGAPP_NAME:
        writer.install_files(ConfigArgs.PKGAPP_NAME, "install")
    if ConfigArgs.PKGUPD_NAME:
        writer.install_files(ConfigArgs.PKGUPD_NAME, "install -k updater.key")

    # Save files
    if ConfigArgs.FILE_NAME:
        print(">>> Save files ...")
        writer.save_files(ConfigArgs.FILE_NAME)

    # Set auto boot
    if not ConfigArgs.NO_SET_BOOTABLE:
        print(">>> Save Configuration to FlashROM ...")
        writer.send("set bootable M0P")
        writer.wait_for_prompt()

    # Sync all cached data to flash
    writer.send("sync")
    writer.wait_for_prompt()

    if REBOOT_AT_END:
        print("Restarting the board ...")
        writer.send("reboot")

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        print("Canceled by keyboard interrupt.")
        pass



l = [1, 2, 3, 4]
s = set(l)
outs = [s.difference(),
        s.difference({1}),
        s.difference({1}, [1, 2]),
        s.difference({1}, {1, 2}, {2, 3})]
for out in outs:
    print(sorted(out))

s = set(l)
print(s.difference_update())
print(sorted(s))
print(s.difference_update({1}))
print(sorted(s))
print(s.difference_update({1}, [2]))
print(sorted(s))

s.difference_update(s)
print(s)



for n in range(20):
    print('testing dict with {} items'.format(n))
    for i in range(n):
        # create dict
        d = dict()
        for j in range(n):
            d[str(j)] = j
        print(len(d))

        # delete an item
        del d[str(i)]
        print(len(d))

        # check items
        for j in range(n):
            if str(j) in d:
                if j == i:
                    print(j, 'in d, but it should not be')
            else:
                if j != i:
                    print(j, 'not in d, but it should be')



try:
    from collections import namedtuple
except ImportError:
    print("SKIP")
    raise SystemExit

T = namedtuple("Tup", ["foo", "bar"])
# CPython prints fully qualified name, what we don't bother to do so far
#print(T)
for t in T(1, 2), T(bar=1, foo=2):
    print(t)
    print(t[0], t[1])
    print(t.foo, t.bar)

    print(len(t))
    print(bool(t))
    print(t + t)
    print(t * 3)

    print([f for f in t])

    print(isinstance(t, tuple))

    # Check tuple can compare equal to namedtuple with same elements
    print(t == (t[0], t[1]), (t[0], t[1]) == t)

# Create using positional and keyword args
print(T(3, bar=4))

try:
    t[0] = 200
except TypeError:
    print("TypeError")
try:
    t.bar = 200
except AttributeError:
    print("AttributeError")

try:
    t = T(1)
except TypeError:
    print("TypeError")

try:
    t = T(1, 2, 3)
except TypeError:
    print("TypeError")

try:
    t = T(foo=1)
except TypeError:
    print("TypeError")

try:
    t = T(1, foo=1)
except TypeError:
    print("TypeError")

# enough args, but kw is wrong
try:
    t = T(1, baz=3)
except TypeError:
    print("TypeError")

# bad argument for member spec
try:
    namedtuple('T', 1)
except TypeError:
    print("TypeError")

# Try tuple
T4 = namedtuple("TupTuple", ("foo", "bar"))
t = T4(1, 2)
print(t.foo, t.bar)

# Try single string with comma field separator
# Not implemented so far
#T2 = namedtuple("TupComma", "foo,bar")
#t = T2(1, 2)



# tests int constant folding in compiler

# negation
print(-0x3fffffff) # 32-bit edge case
print(-0x3fffffffffffffff) # 64-bit edge case
print(-(-0x3fffffff - 1)) # 32-bit edge case
print(-(-0x3fffffffffffffff - 1)) # 64-bit edge case

# 1's complement
print(~0x3fffffff) # 32-bit edge case
print(~0x3fffffffffffffff) # 64-bit edge case
print(~(-0x3fffffff - 1)) # 32-bit edge case
print(~(-0x3fffffffffffffff - 1)) # 64-bit edge case

# zero big-num on rhs
print(1 + ((1 << 65) - (1 << 65)))

# negative big-num on rhs
print(1 + (-(1 << 65)))



print("".isspace())
print(" \t\n\r\v\f".isspace())
print("a".isspace())
print(" \t\n\r\v\fa".isspace())
print("".isalpha())
print("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".isalpha())
print("0abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".isalpha())
print("this ".isalpha())
print("".isdigit())
print("0123456789".isdigit())
print("0123456789a".isdigit())
print("0123456789 ".isdigit())
print("".isupper())
print("CHEESE-CAKE WITH ... _FROSTING_*99".isupper())
print("aB".isupper())
print("".islower())
print("cheese-cake with ... _frosting_*99".islower())
print("aB".islower())
print("123".islower())
print("123a".islower())



import sys

try:
    sys.settrace
except AttributeError:
    print("SKIP")
    raise SystemExit


def print_stacktrace(frame, level=0):
    # Ignore CPython specific helpers.
    if frame.f_globals["__name__"].find("importlib") != -1:
        print_stacktrace(frame.f_back, level)
        return

    print(
        "%2d: %s@%s:%s => %s:%d"
        % (
            level,
            "  ",
            frame.f_globals["__name__"],
            frame.f_code.co_name,
            # Keep just the filename.
            "sys_settrace_" + frame.f_code.co_filename.split("sys_settrace_")[-1],
            frame.f_lineno,
        )
    )

    if frame.f_back:
        print_stacktrace(frame.f_back, level + 1)


class _Prof:
    trace_count = 0

    def trace_tick(self, frame, event, arg):
        self.trace_count += 1
        print_stacktrace(frame)


__prof__ = _Prof()

alice_handler_set = False


def trace_tick_handler_alice(frame, event, arg):
    print("### trace_handler::Alice event:", event)
    __prof__.trace_tick(frame, event, arg)
    return trace_tick_handler_alice


bob_handler_set = False


def trace_tick_handler_bob(frame, event, arg):
    print("### trace_handler::Bob event:", event)
    __prof__.trace_tick(frame, event, arg)
    return trace_tick_handler_bob


def trace_tick_handler(frame, event, arg):
    # Ignore CPython specific helpers.
    to_ignore = ["importlib", "zipimport", "encodings"]
    frame_name = frame.f_globals["__name__"]
    if any(name in frame_name for name in to_ignore):
        return

    print("### trace_handler::main event:", event)
    __prof__.trace_tick(frame, event, arg)

    if frame.f_code.co_name != "factorial":
        return trace_tick_handler

    global alice_handler_set
    if event == "call" and not alice_handler_set:
        alice_handler_set = True
        return trace_tick_handler_alice

    global bob_handler_set
    if event == "call" and not bob_handler_set:
        bob_handler_set = True
        return trace_tick_handler_bob

    return trace_tick_handler


def factorial(n):
    if n == 0:
        return 1
    else:
        return n * factorial(n - 1)


def do_tests():
    # These commands are here to demonstrate some execution being traced.
    print("Who loves the sun?")
    print("Not every-", factorial(3))

    from sys_settrace_subdir import sys_settrace_generic

    sys_settrace_generic.run_tests()
    return


sys.settrace(trace_tick_handler)
do_tests()
sys.settrace(None)

print("\n------------------ script exited ------------------")
print("Total traces executed: ", __prof__.trace_count)



# test loading constants in viper functions


@micropython.viper
def f():
    return b"bytes"


print(f())


@micropython.viper
def f():
    @micropython.viper
    def g() -> int:
        return 123

    return g


print(f()())



# test passing arguments


@micropython.asm_thumb
def arg0():
    mov(r0, 1)


print(arg0())


@micropython.asm_thumb
def arg1(r0):
    add(r0, r0, 1)


print(arg1(1))


@micropython.asm_thumb
def arg2(r0, r1):
    add(r0, r0, r1)


print(arg2(1, 2))


@micropython.asm_thumb
def arg3(r0, r1, r2):
    add(r0, r0, r1)
    add(r0, r0, r2)


print(arg3(1, 2, 3))


@micropython.asm_thumb
def arg4(r0, r1, r2, r3):
    add(r0, r0, r1)
    add(r0, r0, r2)
    add(r0, r0, r3)


print(arg4(1, 2, 3, 4))



# Test builtin slice indices resolution

# A class that returns an item key
class A:
    def __getitem__(self, idx):
        return idx

# Make sure that we have slices and .indices()
try:
    A()[2:5].indices(10)
except:
    print("SKIP")
    raise SystemExit

print(A()[:].indices(10))
print(A()[2:].indices(10))
print(A()[:7].indices(10))
print(A()[2:7].indices(10))
print(A()[2:7:2].indices(10))
print(A()[2:7:-2].indices(10))
print(A()[7:2:2].indices(10))
print(A()[7:2:-2].indices(10))

print(A()[2:7:2].indices(5))
print(A()[2:7:-2].indices(5))
print(A()[7:2:2].indices(5))
print(A()[7:2:-2].indices(5))



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2014 MicroPython & CircuitPython contributors (https://github.com/adafruit/circuitpython/graphs/contributors)
#
# SPDX-License-Identifier: MIT

import argparse
import os
import os.path
from pathlib import Path
import semver
import subprocess

# Compatible with Python 3.4 due to travis using trusty as default.


def version_string(path=None, *, valid_semver=False):
    version = None
    try:
        tag = subprocess.check_output("git describe --tags --exact-match", shell=True, cwd=path)
        version = tag.strip().decode("utf-8", "strict")
    except subprocess.CalledProcessError:
        describe = subprocess.check_output("git describe --tags", shell=True, cwd=path)
        tag, additional_commits, commitish = (
            describe.strip().decode("utf-8", "strict").rsplit("-", maxsplit=2)
        )
        commitish = commitish[1:]
        if valid_semver:
            version_info = semver.parse_version_info(tag)
            if not version_info.prerelease:
                version = (
                    semver.bump_patch(tag)
                    + "-alpha.0.plus."
                    + additional_commits
                    + "+"
                    + commitish
                )
            else:
                version = tag + ".plus." + additional_commits + "+" + commitish
        else:
            version = commitish
    return version


# Visit all the .py files in topdir. Replace any __version__ = "0.0.0-auto.0" type of info
# with actual version info derived from git.
def copy_and_process(in_dir, out_dir):
    for root, subdirs, files in os.walk(in_dir):

        # Skip library examples directory and subfolders.
        relative_path_parts = Path(root).relative_to(in_dir).parts
        if relative_path_parts and relative_path_parts[0] in ["examples", "docs", "tests"]:
            del subdirs[:]
            continue

        for file in files:
            # Skip top-level setup.py (module install info) and conf.py (sphinx config),
            # which are not part of the library
            if (root == in_dir) and file in ("conf.py", "setup.py"):
                continue

            input_file_path = Path(root, file)
            output_file_path = Path(out_dir, input_file_path.relative_to(in_dir))

            if file.endswith(".py"):
                if not output_file_path.parent.exists():
                    output_file_path.parent.mkdir(parents=True)
                with input_file_path.open("r") as input, output_file_path.open("w") as output:
                    for line in input:
                        if line.startswith("__version__"):
                            module_version = version_string(root, valid_semver=True)
                            line = line.replace("0.0.0-auto.0", module_version)
                        output.write(line)


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="""\
    Copy and pre-process .py files into output directory, before freezing.
    1. Remove top-level repo directory.
    2. Update __version__ info.
    3. Remove examples.
    4. Remove non-library setup.py and conf.py"""
    )
    argparser.add_argument(
        "in_dirs",
        metavar="input-dir",
        nargs="+",
        help="top-level code dirs (may be git repo dirs)",
    )
    argparser.add_argument("-o", "--out_dir", help="output directory")
    args = argparser.parse_args()

    for in_dir in args.in_dirs:
        copy_and_process(in_dir, args.out_dir)



# default separator (whitespace)
print("a b".split())
print("   a   b    ".split(None))
print("   a   b    ".split(None, 1))
print("   a   b    ".split(None, 2))
print("   a   b  c  ".split(None, 1))
print("   a   b  c  ".split(None, 0))
print("   a   b  c  ".split(None, -1))
print("foo\n\t\x07\v\nbar".split())
print("foo\nbar\n".split())

# empty separator should fail
try:
    "abc".split('')
except ValueError:
    print("ValueError")

# non-empty separator
print("abc".split("a"))
print("abc".split("b"))
print("abc".split("c"))
print("abc".split("z"))
print("abc".split("ab"))
print("abc".split("bc"))
print("abc".split("abc"))
print("abc".split("abcd"))
print("abcabc".split("bc"))
print("abcabc".split("bc", 0))
print("abcabc".split("bc", 1))
print("abcabc".split("bc", 2))

print(b"abcabc".split(b"bc", 2))



# test construction of bytes from array with float type

try:
    from uarray import array
except ImportError:
    try:
        from array import array
    except ImportError:
        print("SKIP")
        raise SystemExit

print(bytes(array("f", [1, 2.5])))



try:
    import uhashlib as hashlib
except ImportError:
    try:
        import hashlib
    except ImportError:
        # This is neither uPy, nor cPy, so must be uPy with
        # uhashlib module disabled.
        print("SKIP")
        raise SystemExit


h = hashlib.sha256()
print(h.digest())

h = hashlib.sha256()
h.update(b"123")
print(h.digest())

h = hashlib.sha256()
h.update(b"abcd" * 1000)
print(h.digest())

print(hashlib.sha256(b"\xff" * 64).digest())

# 56 bytes is a boundary case in the algorithm
print(hashlib.sha256(b"\xff" * 56).digest())

sha256 = hashlib.sha256(b"hello")
try:
    sha256.update(u"world")
except TypeError as e:
    print("TypeError")
print(sha256.digest())

# TODO: running .digest() several times in row is not supported()
# h = hashlib.sha256(b'123')
# print(h.digest())
# print(h.digest())

# TODO: partial digests are not supported
# h = hashlib.sha256(b'123')
# print(h.digest())
# h.update(b'456')
# print(h.digest())



"""
categories: Modules,builtins
description: Second argument to next() is not implemented
cause: MicroPython is optimised for code space.
workaround: Instead of ``val = next(it, deflt)`` use::

    try:
        val = next(it)
    except StopIteration:
        val = deflt
"""
print(next(iter(range(0)), 42))



try:
    str.partition
except AttributeError:
    print("SKIP")
    raise SystemExit

print("asdf".partition('g'))
print("asdf".partition('a'))
print("asdf".partition('s'))
print("asdf".partition('f'))
print("asdf".partition('d'))
print("asdf".partition('asd'))
print("asdf".partition('sdf'))
print("asdf".partition('as'))
print("asdf".partition('df'))
print("asdf".partition('asdf'))
print("asdf".partition('asdfa'))
print("asdf".partition('fasdf'))
print("asdf".partition('fasdfa'))
print("abba".partition('a'))
print("abba".partition('b'))

try:
    print("asdf".partition(1))
except TypeError:
    print("Raised TypeError")
else:
    print("Did not raise TypeError")

try:
    print("asdf".partition(''))
except ValueError:
    print("Raised ValueError")
else:
    print("Did not raise ValueError")

# Bytes
print(b"abba".partition(b'b'))
try:
    print(b"abba".partition('b'))
except TypeError:
    print("Raised TypeError")
try:
    print("abba".partition(b'b'))
except TypeError:
    print("Raised TypeError")



@micropython.asm_thumb
def sdiv(r0, r1):
    sdiv(r0, r0, r1)


@micropython.asm_thumb
def udiv(r0, r1):
    udiv(r0, r0, r1)


print(sdiv(1234, 3))
print(sdiv(-1234, 3))
print(sdiv(1234, -3))
print(sdiv(-1234, -3))

print(udiv(1234, 3))
print(udiv(0xFFFFFFFF, 0x7FFFFFFF))
print(udiv(0xFFFFFFFF, 0xFFFFFFFF))



try:
    import uerrno
    import uos
except ImportError:
    print("SKIP")
    raise SystemExit

try:
    uos.VfsFat
except AttributeError:
    print("SKIP")
    raise SystemExit


class RAMFS:

    SEC_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.SEC_SIZE)

    def readblocks(self, n, buf):
        # print("readblocks(%s, %x(%d))" % (n, id(buf), len(buf)))
        for i in range(len(buf)):
            buf[i] = self.data[n * self.SEC_SIZE + i]
        return 0

    def writeblocks(self, n, buf):
        # print("writeblocks(%s, %x)" % (n, id(buf)))
        for i in range(len(buf)):
            self.data[n * self.SEC_SIZE + i] = buf[i]
        return 0

    def ioctl(self, op, arg):
        # print("ioctl(%d, %r)" % (op, arg))
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return len(self.data) // self.SEC_SIZE
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.SEC_SIZE


try:
    bdev = RAMFS(50)
except MemoryError:
    print("SKIP")
    raise SystemExit

uos.VfsFat.mkfs(bdev)
vfs = uos.VfsFat(bdev)
uos.mount(vfs, "/ramdisk")
uos.chdir("/ramdisk")

try:
    vfs.mkdir("foo_dir")
except OSError as e:
    print(e.errno == uerrno.EEXIST)

try:
    vfs.remove("foo_dir")
except OSError as e:
    print(e.errno == uerrno.EISDIR)

try:
    vfs.remove("no_file.txt")
except OSError as e:
    print(e.errno == uerrno.ENOENT)

try:
    vfs.rename("foo_dir", "/null/file")
except OSError as e:
    print(e.errno == uerrno.ENOENT)

# file in dir
with open("foo_dir/file-in-dir.txt", "w+t") as f:
    f.write("data in file")

with open("foo_dir/file-in-dir.txt", "r+b") as f:
    print(f.read())

with open("foo_dir/sub_file.txt", "w") as f:
    f.write("subdir file")

# directory not empty
try:
    vfs.rmdir("foo_dir")
except OSError as e:
    print(e.errno == uerrno.EACCES)

# trim full path
vfs.rename("foo_dir/file-in-dir.txt", "foo_dir/file.txt")
print(list(vfs.ilistdir("foo_dir")))

vfs.rename("foo_dir/file.txt", "moved-to-root.txt")
print(list(vfs.ilistdir()))

# check that renaming to existing file will overwrite it
with open("temp", "w") as f:
    f.write("new text")
vfs.rename("temp", "moved-to-root.txt")
print(list(vfs.ilistdir()))
with open("moved-to-root.txt") as f:
    print(f.read())

# valid removes
vfs.remove("foo_dir/sub_file.txt")
vfs.rmdir("foo_dir")
print(list(vfs.ilistdir()))

# disk full
try:
    bsize = vfs.statvfs("/ramdisk")[0]
    free = vfs.statvfs("/ramdisk")[2] + 1
    f = open("large_file.txt", "wb")
    f.write(bytearray(bsize * free))
except OSError as e:
    print("ENOSPC:", e.errno == 28)  # uerrno.ENOSPC
f.close()



# test arithmetic operators


@micropython.viper
def add(x: int, y: int):
    print(x + y)
    print(y + x)


add(1, 2)
add(42, 3)
add(-1, 2)
add(-42, -3)


@micropython.viper
def sub(x: int, y: int):
    print(x - y)
    print(y - x)


sub(1, 2)
sub(42, 3)
sub(-1, 2)
sub(-42, -3)


@micropython.viper
def mul(x: int, y: int):
    print(x * y)
    print(y * x)


mul(0, 1)
mul(1, -1)
mul(1, 2)
mul(8, 3)
mul(-3, 4)
mul(-9, -6)


@micropython.viper
def shl(x: int, y: int):
    print(x << y)


shl(1, 0)
shl(1, 3)
shl(1, 30)
shl(42, 10)
shl(-42, 10)


@micropython.viper
def shr(x: int, y: int):
    print(x >> y)


shr(1, 0)
shr(1, 3)
shr(42, 2)
shr(-42, 2)


@micropython.viper
def and_(x: int, y: int):
    print(x & y, y & x)


and_(1, 0)
and_(1, 3)
and_(0xF0, 0x3F)
and_(-42, 6)


@micropython.viper
def or_(x: int, y: int):
    print(x | y, y | x)


or_(1, 0)
or_(1, 2)
or_(-42, 5)


@micropython.viper
def xor(x: int, y: int):
    print(x ^ y, y ^ x)


xor(1, 0)
xor(1, 2)
xor(-42, 5)



# Source: https://github.com/python/pyperformance
# License: MIT

# Artificial, floating point-heavy benchmark originally used by Factor.

from math import sin, cos, sqrt


class Point(object):
    __slots__ = ("x", "y", "z")

    def __init__(self, i):
        self.x = x = sin(i)
        self.y = cos(i) * 3
        self.z = (x * x) / 2

    def __repr__(self):
        return "<Point: x=%s, y=%s, z=%s>" % (self.x, self.y, self.z)

    def normalize(self):
        x = self.x
        y = self.y
        z = self.z
        norm = sqrt(x * x + y * y + z * z)
        self.x /= norm
        self.y /= norm
        self.z /= norm

    def maximize(self, other):
        self.x = self.x if self.x > other.x else other.x
        self.y = self.y if self.y > other.y else other.y
        self.z = self.z if self.z > other.z else other.z
        return self


def maximize(points):
    next = points[0]
    for p in points[1:]:
        next = next.maximize(p)
    return next


def benchmark(n):
    points = [None] * n
    for i in range(n):
        points[i] = Point(i)
    for p in points:
        p.normalize()
    return maximize(points)


###########################################################################
# Benchmark interface

bm_params = {
    (50, 25): (1, 150),
    (100, 100): (1, 250),
    (1000, 1000): (10, 1500),
    (5000, 1000): (20, 3000),
}


def bm_setup(params):
    state = None

    def run():
        nonlocal state
        for _ in range(params[0]):
            state = benchmark(params[1])

    def result():
        return params[0] * params[1], "Point(%.4f, %.4f, %.4f)" % (state.x, state.y, state.z)

    return run, result



# test builtin exec

try:
    exec
except NameError:
    print("SKIP")
    raise SystemExit

print(exec("def foo(): return 42"))
print(foo())

d = {}
exec("def bar(): return 84", d)
print(d["bar"]())

# passing None/dict as args to globals/locals
foo = 11
exec('print(foo)')
exec('print(foo)', None)
exec('print(foo)', {'foo':3}, None)
exec('print(foo)', None, {'foo':3})
exec('print(foo)', None, {'bar':3})
exec('print(foo)', {'bar':3}, locals())

try:
    exec('print(foo)', {'bar':3}, None)
except NameError:
    print('NameError')

# invalid arg passed to globals
try:
    exec('print(1)', 'foo')
except TypeError:
    print('TypeError')

# invalid arg passed to locals
try:
    exec('print(1)', None, 123)
except TypeError:
    print('TypeError')



@micropython.viper
def f2a(x, y):
    return x


@micropython.native
def call(r):
    f = f2a
    for _ in r:
        f(1, 2)


bm_params = {
    (50, 10): (15000,),
    (100, 10): (30000,),
    (1000, 10): (300000,),
    (5000, 10): (1500000,),
}


def bm_setup(params):
    return lambda: call(range(params[0])), lambda: (params[0] // 1000, None)



# test construction of array.array from different objects

try:
    from uarray import array
except ImportError:
    try:
        from array import array
    except ImportError:
        print("SKIP")
        raise SystemExit

# raw copy from bytes, bytearray
print(array('h', b'12'))



# test loading from bytes and bytearray (introduced in Python 3.6)

try:
    import ujson as json
except ImportError:
    try:
        import json
    except ImportError:
        print("SKIP")
        raise SystemExit

print(json.loads(b"[1,2]"))
print(json.loads(bytearray(b"[null]")))



# test that __getattr__ and instance members don't override builtins
class C:
    def __init__(self):
        self.__add__ = lambda: print('member __add__')
    def __add__(self, x):
        print('__add__')
    def __getattr__(self, attr):
        print('__getattr__', attr)
        return None

c = C()
c.add # should call __getattr__
c.__add__() # should load __add__ instance directly
c + 1 # should call __add__ method directly



# Test that we can raise and catch (preallocated) exception
# without memory allocation.
import micropython

e = ValueError("error")


def func():
    micropython.heap_lock()
    try:
        # This works as is because traceback is not allocated
        # if not possible (heap is locked, no memory). If heap
        # is not locked, this would allocate a traceback entry.
        # To avoid that, traceback should be warmed up (by raising
        # it once after creation) and then cleared before each
        # raise with:
        # e.__traceback__ = None
        raise e
    except Exception as e2:
        print(e2)
    micropython.heap_unlock()


func()
print("ok")



# Test the Task.done() method

try:
    import uasyncio as asyncio
except ImportError:
    try:
        import asyncio
    except ImportError:
        print("SKIP")
        raise SystemExit


async def task(t, exc=None):
    print("task start")
    if t >= 0:
        await asyncio.sleep(t)
    if exc:
        raise exc
    print("task done")


async def main():
    # Task that finishes immediately.
    print("=" * 10)
    t = asyncio.create_task(task(-1))
    print(t.done())
    await asyncio.sleep(0)
    print(t.done())
    await t
    print(t.done())

    # Task that starts, runs and finishes.
    print("=" * 10)
    t = asyncio.create_task(task(0.01))
    print(t.done())
    await asyncio.sleep(0)
    print(t.done())
    await t
    print(t.done())

    # Task that raises immediately.
    print("=" * 10)
    t = asyncio.create_task(task(-1, ValueError))
    print(t.done())
    await asyncio.sleep(0)
    print(t.done())
    try:
        await t
    except ValueError as er:
        print(repr(er))
    print(t.done())

    # Task that raises after a delay.
    print("=" * 10)
    t = asyncio.create_task(task(0.01, ValueError))
    print(t.done())
    await asyncio.sleep(0)
    print(t.done())
    try:
        await t
    except ValueError as er:
        print(repr(er))
    print(t.done())


asyncio.run(main())



import uio
import sys

try:
    uio.resource_stream
except AttributeError:
    print("SKIP")
    raise SystemExit

buf = uio.resource_stream("data", "file2")
print(buf.read())

# resource_stream(None, ...) look ups from current dir, hence sys.path[0] hack
buf = uio.resource_stream(None, sys.path[0] + "/data/file2")
print(buf.read())



# The aim with this test is to hit the maximum resize/rehash size of a dict,
# where there are no more primes in the table of growing allocation sizes.
# This value is 54907 items.

d = {}
try:
    for i in range(54908):
        d[i] = i
except MemoryError:
    pass

# Just check the dict still has the first value
print(d[0])



print(b"".replace(b"a", b"b"))
print(b"aaa".replace(b"a", b"b", 0))
print(b"aaa".replace(b"a", b"b", -5))
print(b"asdfasdf".replace(b"a", b"b"))
print(b"aabbaabbaabbaa".replace(b"aa", b"cc", 3))
print(b"a".replace(b"aa", b"bb"))
print(b"testingtesting".replace(b"ing", b""))
print(b"testINGtesting".replace(b"ing", b"ING!"))

print(b"".replace(b"", b"1"))
print(b"A".replace(b"", b"1"))
print(b"AB".replace(b"", b"1"))
print(b"AB".replace(b"", b"12"))



# test comparison operators
@micropython.viper
def f(x: int, y: int):
    if x < y:
        print(x, "<", y)
    if x > y:
        print(x, ">", y)
    if x == y:
        print(x, "==", y)
    if x <= y:
        print(x, "<=", y)
    if x >= y:
        print(x, ">=", y)
    if x != y:
        print(x, "!=", y)


f(1, 1)
f(2, 1)
f(1, 2)
f(2, -1)
f(-2, 1)



"""
categories: Types,str
description: str.ljust() and str.rjust() not implemented
cause: MicroPython is highly optimized for memory usage. Easy workarounds available.
workaround: Instead of ``s.ljust(10)`` use ``"%-10s" % s``, instead of ``s.rjust(10)`` use ``"% 10s" % s``. Alternatively, ``"{:<10}".format(s)`` or ``"{:>10}".format(s)``.
"""
print("abc".ljust(10))



# basic class

def go():
    class C:
        def f():
            print(1)

        def g(self):
            print(2)

        def set(self, value):
            self.value = value

        def print(self):
            print(self.value)

    C.f()
    C()
    C().g()

    o = C()
    o.set(3)
    o.print()

    C.set(o, 4)
    C.print(o)

go()



# test array + array
try:
    import uarray as array
except ImportError:
    try:
        import array
    except ImportError:
        print("SKIP")
        raise SystemExit

a1 = array.array('I', [1])
a2 = array.array('I', [2])
print(a1 + a2)

a1 += array.array('I', [3, 4])
print(a1)

a1.extend(array.array('I', [5]))
print(a1)



try:
    frozenset
except NameError:
    print("SKIP")
    raise SystemExit

s = frozenset({1, 2, 3, 4})
try:
    print(s.add(5))
except AttributeError:
    print("AttributeError")

try:
    print(s.update([5]))
except AttributeError:
    print("AttributeError")



# test __delattr__ and __setattr__

# feature test for __setattr__/__delattr__
try:
    class Test():
        def __delattr__(self, attr): pass
    del Test().noexist
except AttributeError:
    print('SKIP')
    raise SystemExit

# this class just prints the calls to see if they were executed
class A():
    def __getattr__(self, attr):
        print('get', attr)
        return 1
    def __setattr__(self, attr, val):
        print('set', attr, val)
    def __delattr__(self, attr):
        print('del', attr)
a = A()

# check basic behaviour
print(getattr(a, 'foo'))
setattr(a, 'bar', 2)
delattr(a, 'baz')

# check meta behaviour
getattr(a, '__getattr__') # should not call A.__getattr__
getattr(a, '__setattr__') # should not call A.__getattr__
getattr(a, '__delattr__') # should not call A.__getattr__
setattr(a, '__setattr__', 1) # should call A.__setattr__
delattr(a, '__delattr__') # should call A.__delattr__

# this class acts like a dictionary
class B:
    def __init__(self, d):
        # store the dict in the class, not instance, so
        # we don't get infinite recursion in __getattr_
        B.d = d

    def __getattr__(self, attr):
        if attr in B.d:
            return B.d[attr]
        else:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        B.d[attr] = value

    def __delattr__(self, attr):
        del B.d[attr]

a = B({"a":1, "b":2})
print(a.a, a.b)
a.a = 3
print(a.a, a.b)
del a.a
try:
    print(a.a)
except AttributeError:
    print("AttributeError")

# test object.__setattr__
class C:
    def __init__(self):
        pass

    def __setattr__(self, attr, value):
        print(attr, "=", value)

    def __delattr__(self, attr):
        print("del", attr)

c = C()
c.a = 5
try:
    print(c.a)
except AttributeError:
    print("AttributeError")

object.__setattr__(c, "a", 5)
super(C, c).__setattr__("b", 6)
print(c.a)
print(c.b)

try:
    # attribute name must be string
    object.__setattr__(c, 5, 5)
except TypeError:
    print("TypeError")


# test object.__delattr__
del c.a
print(c.a)

object.__delattr__(c, "a")
try:
    print(c.a)
except AttributeError:
    print("AttributeError")

super(C, c).__delattr__("b")
try:
    print(c.b)
except AttributeError:
    print("AttributeError")

try:
    object.__delattr__(c, "c")
except AttributeError:
    print("AttributeError")



# function with just varargs
def f1(*args):
    print(args)

f1()
f1(1)
f1(1, 2)

# function with 1 arg, then varargs
def f2(a, *args):
    print(a, args)

f2(1)
f2(1, 2)
f2(1, 2, 3)

# function with 2 args, then varargs
def f3(a, b, *args):
    print(a, b, args)

f3(1, 2)
f3(1, 2, 3)
f3(1, 2, 3, 4)

# function with 1 default arg, then varargs
def f4(a=0, *args):
    print(a, args)

f4()
f4(1)
f4(1, 2)
f4(1, 2, 3)

# function with 1 arg, 1 default arg, then varargs
def f5(a, b=0, *args):
    print(a, b, args)

f5(1)
f5(1, 2)
f5(1, 2, 3)
f5(1, 2, 3, 4)



# test for-else statement

# test optimised range with simple else
for i in range(2):
    print(i)
else:
    print('else')

# test optimised range with break over else
for i in range(2):
    print(i)
    break
else:
    print('else')

# test nested optimised range with continue in the else
for i in range(4):
    print(i)
    for j in range(4):
        pass
    else:
        continue
    break

# test optimised range with non-constant end value
N = 2
for i in range(N):
    print(i)
else:
    print('else')

# test generic iterator with simple else
for i in [0, 1]:
    print(i)
else:
    print('else')

# test generic iterator with break over else
for i in [0, 1]:
    print(i)
    break
else:
    print('else')



try:
    import framebuf
except ImportError:
    print("SKIP")
    raise SystemExit

w = 5
h = 16
size = w * h // 8
buf = bytearray(size)
maps = {
    framebuf.MONO_VLSB: "MONO_VLSB",
    framebuf.MONO_HLSB: "MONO_HLSB",
    framebuf.MONO_HMSB: "MONO_HMSB",
}

for mapping in maps.keys():
    for x in range(size):
        buf[x] = 0
    fbuf = framebuf.FrameBuffer(buf, w, h, mapping)
    print(maps[mapping])
    # access as buffer
    print(memoryview(fbuf)[0])

    # fill
    fbuf.fill(1)
    print(buf)
    fbuf.fill(0)
    print(buf)

    # put pixel
    fbuf.pixel(0, 0, 1)
    fbuf.pixel(4, 0, 1)
    fbuf.pixel(0, 15, 1)
    fbuf.pixel(4, 15, 1)
    print(buf)

    # clear pixel
    fbuf.pixel(4, 15, 0)
    print(buf)

    # get pixel
    print(fbuf.pixel(0, 0), fbuf.pixel(1, 1))

    # hline
    fbuf.fill(0)
    fbuf.hline(0, 1, w, 1)
    print("hline", buf)

    # vline
    fbuf.fill(0)
    fbuf.vline(1, 0, h, 1)
    print("vline", buf)

    # rect
    fbuf.fill(0)
    fbuf.rect(1, 1, 3, 3, 1)
    print("rect", buf)

    # fill rect
    fbuf.fill(0)
    fbuf.fill_rect(0, 0, 0, 3, 1)  # zero width, no-operation
    fbuf.fill_rect(1, 1, 3, 3, 1)
    print("fill_rect", buf)

    # line
    fbuf.fill(0)
    fbuf.line(1, 1, 3, 3, 1)
    print("line", buf)

    # line steep negative gradient
    fbuf.fill(0)
    fbuf.line(3, 3, 2, 1, 1)
    print("line", buf)

    # scroll
    fbuf.fill(0)
    fbuf.pixel(2, 7, 1)
    fbuf.scroll(0, 1)
    print(buf)
    fbuf.scroll(0, -2)
    print(buf)
    fbuf.scroll(1, 0)
    print(buf)
    fbuf.scroll(-1, 0)
    print(buf)
    fbuf.scroll(2, 2)
    print(buf)

    # print text
    fbuf.fill(0)
    fbuf.text("hello", 0, 0, 1)
    print(buf)
    fbuf.text("hello", 0, 0, 0)  # clear
    print(buf)

    # char out of font range set to chr(127)
    fbuf.text(str(chr(31)), 0, 0)
    print(buf)
    print()

# test invalid constructor, and stride argument
try:
    fbuf = framebuf.FrameBuffer(buf, w, h, -1, w)
except ValueError:
    print("ValueError")

# test legacy constructor
fbuf = framebuf.FrameBuffer1(buf, w, h)
fbuf = framebuf.FrameBuffer1(buf, w, h, w)
print(framebuf.MVLSB == framebuf.MONO_VLSB)



try:
    import utime as time
except ImportError:
    import time

DAYS_PER_MONTH = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

tzseconds = -time.mktime((1970, 1, 1, 14, 0, 0, 0, 0, 0))


def is_leap(year):
    return (year % 4) == 0


def test():
    seconds = 0
    wday = 3  # Jan 1, 1970 was a Thursday
    for year in range(1970, 2038):
        print("Testing %d" % year)
        yday = 1
        for month in range(1, 13):
            if month == 2 and is_leap(year):
                DAYS_PER_MONTH[2] = 29
            else:
                DAYS_PER_MONTH[2] = 28
            for day in range(1, DAYS_PER_MONTH[month] + 1):
                secs = time.mktime((year, month, day, 14, 0, 0, 0, 0, 0)) + tzseconds
                if secs != seconds:
                    print(
                        "mktime failed for %d-%02d-%02d got %d expected %d"
                        % (year, month, day, secs, seconds)
                    )
                    return
                tuple = time.localtime(seconds)
                secs = time.mktime(tuple)
                if secs != seconds:
                    print(
                        "localtime failed for %d-%02d-%02d got %d expected %d"
                        % (year, month, day, secs, seconds)
                    )
                    return
                seconds += 86400
                if yday != tuple[7]:
                    print(
                        "locatime for %d-%02d-%02d got yday %d, expecting %d"
                        % (year, month, day, tuple[7], yday)
                    )
                    return
                if wday != tuple[6]:
                    print(
                        "locatime for %d-%02d-%02d got wday %d, expecting %d"
                        % (year, month, day, tuple[6], wday)
                    )
                    return
                yday += 1
                wday = (wday + 1) % 7


test()



# Tests against https://www.python.org/dev/peps/pep-0498/

assert f'no interpolation' == 'no interpolation'
assert f"no interpolation" == 'no interpolation'

# Quoth the PEP:
# Backslashes may not appear anywhere within expressions. Comments, using the
# '#' character, are not allowed inside an expression
#
# CPython (3.7.4 on Linux) raises a SyntaxError here:
# >>> f'{#}'
#   File "<stdin>", line 1
# SyntaxError: f-string expression part cannot include '#'
# >>> f'{\}'
#   File "<stdin>", line 1
# SyntaxError: f-string expression part cannot include a backslash
# >>> f'{\\}'
#   File "<stdin>", line 1
# SyntaxError: f-string expression part cannot include a backslash
# >>> f'{\#}'
#   File "<stdin>", line 1
# SyntaxError: f-string expression part cannot include a backslash

# Backslashes and comments allowed outside expression
assert f"\\" == "\\"
assert f'#' == '#'

## But not inside
try:
    eval("f'{\}'")
except SyntaxError:
    pass
else:
    raise AssertionError('f-string with backslash in expression did not raise SyntaxError')

try:
    eval("f'{#}'")
except SyntaxError:
    pass
else:
    raise AssertionError('f-string with \'#\' in expression did not raise SyntaxError')

# Quoth the PEP:
# While scanning the string for expressions, any doubled braces '{{' or '}}'
# inside literal portions of an f-string are replaced by the corresponding
# single brace.  Doubled literal opening braces do not signify the start of an
# expression. A single closing curly brace '}' in the literal portion of a
# string is an error: literal closing curly braces must be doubled '}}' in
# order to represent a single closing brace.
#
# CPython (3.7.4 on Linux) raises a SyntaxError for the last case:
# >>> f'{{}'
#   File "<stdin>", line 1
# SyntaxError: f-string: single '}' is not allowed

assert f'{{}}' == '{}'

try:
    eval("f'{{}'")
except ValueError:
    pass
else:
    raise RuntimeError('Expected ValueError for invalid f-string literal bracing')

x = 1
assert f'{x}' == '1'

# Quoth the PEP:
# The expressions that are extracted from the string are evaluated in the
# context where the f-string appeared. This means the expression has full
# access to local and global variables. Any valid Python expression can be
# used, including function and method calls.  Because the f-strings are
# evaluated where the string appears in the source code, there is no additional
# expressiveness available with f-strings. There are also no additional
# security concerns: you could have also just written the same expression, not
# inside of an f-string:

def foo():
    return 20

assert f'result={foo()}' == 'result=20'
assert f'result={foo()}' == 'result={}'.format(foo())
assert f'result={foo()}' == 'result={result}'.format(result=foo())


# Quoth the PEP:
# Adjacent f-strings and regular strings are concatenated. Regular strings are
# concatenated at compile time, and f-strings are concatenated at run time. For
# example, the expression:
#
# >>> x = 10
# >>> y = 'hi'
# >>> 'a' 'b' f'{x}' '{c}' f'str<{y:^4}>' 'd' 'e'
#
# yields the value: 'ab10{c}str< hi >de'
#
# Because strings are concatenated at lexer time rather than parser time in
# MicroPython for mostly RAM efficiency reasons (see
# https://github.com/micropython/micropython/commit/534b7c368dc2af7720f3aaed0c936ef46d773957),
# and because f-strings here are implemented as a syntax translation
# (f'{something}' => '{}'.format(something)), this particular functionality is unimplemented,
# and in the above example, the '{c}' portion will trigger a KeyError on String.format()

x = 10
y = 'hi'
assert (f'h' f'i') == 'hi'
#assert (f'h' 'i') == 'hi'
#assert ('h' f'i') == 'hi'
assert f'{x:^4}' == ' 10 '
#assert ('a' 'b' f'{x}' f'str<{y:^4}>' 'd' 'e') == 'ab10str< hi >de'

# Other tests
assert f'{{{4*10}}}' == '{40}'



# test [0,-0,1,-1] edge cases of bignum

import skip_if
skip_if.no_bigint()

long_zero = (2**64) >> 65
long_neg_zero = -long_zero
long_one = long_zero + 1
long_neg_one = -long_one

cases = [long_zero, long_neg_zero, long_one, long_neg_one]

print(cases)
print([-c for c in cases])
print([~c for c in cases])
print([c >> 1 for c in cases])
print([c << 1 for c in cases])

# comparison of 0/-0/+0
print(long_zero == 0)
print(long_neg_zero == 0)
print(long_one - 1 == 0)
print(long_neg_one + 1 == 0)
print(long_zero < 1)
print(long_zero < -1)
print(long_zero > 1)
print(long_zero > -1)
print(long_neg_zero < 1)
print(long_neg_zero < -1)
print(long_neg_zero > 1)
print(long_neg_zero > -1)



# test builtin slice

# print slice
class A:
    def __getitem__(self, idx):
        print(idx)
        return idx
s = A()[1:2:3]

# check type
print(type(s) is slice)

s = slice(10)
print(s)

s = slice(0, 4)
print(s)

s = slice(0, 4, 2)
print(s)

s = slice(-1)
print(s)
print(s.indices(10))

s = slice(-5, -1)
print(s)
print(s.indices(10))

s = slice(-100, -2, -1)
print(s)
print(s.indices(10))

s = slice(None, None, -2)
print(s)
print(s.indices(10))

s = slice(-100, -2, 0)
print(s)
try:
    print(s.indices(10))
except Exception as e:
    print(e)



# test decorators

def dec(f):
    print('dec')
    return f

def dec_arg(x):
    print(x)
    return lambda f:f

# plain decorator
@dec
def f():
    pass

# decorator with arg
@dec_arg('dec_arg')
def g():
    pass

# decorator of class
@dec
class A:
    pass



"""Supervisor settings"""

from __future__ import annotations

from typing import Optional

runtime: Runtime
"""Runtime information, such as ``runtime.serial_connected``
(USB serial connection status).
This object is the sole instance of `supervisor.Runtime`."""

def enable_autoreload() -> None:
    """Enable autoreload based on USB file write activity."""
    ...

def disable_autoreload() -> None:
    """Disable autoreload based on USB file write activity until
    `enable_autoreload` is called."""
    ...

def set_rgb_status_brightness(brightness: int) -> None:
    """Set brightness of status neopixel from 0-255
    `set_rgb_status_brightness` is called."""
    ...

def reload() -> None:
    """Reload the main Python code and run it (equivalent to hitting Ctrl-D at the REPL)."""
    ...

def set_next_stack_limit(size: int) -> None:
    """Set the size of the stack for the next vm run. If its too large, the default will be used."""
    ...

def set_next_code_file(
    filename: Optional[str],
    *,
    reload_on_success: bool = False,
    reload_on_error: bool = False,
    sticky_on_success: bool = False,
    sticky_on_error: bool = False,
    sticky_on_reload: bool = False
) -> None:
    """Set what file to run on the next vm run.

    When not ``None``, the given ``filename`` is inserted at the front of the usual ['code.py',
    'main.py'] search sequence.

    The optional keyword arguments specify what happens after the specified file has run:

    ``sticky_on_…`` determine whether the newly set filename and options stay in effect: If
    True, further runs will continue to run that file (unless it says otherwise by calling
    ``set_next_code_filename()`` itself). If False, the settings will only affect one run and
    revert to the standard code.py/main.py afterwards.

    ``reload_on_…`` determine how to continue: If False, wait in the usual "Code done running.
    Waiting for reload. / Press any key to enter the REPL. Use CTRL-D to reload." state. If
    True, reload immediately as if CTRL-D was pressed.

    ``…_on_success`` take effect when the program runs to completion or calls ``sys.exit()``.

    ``…_on_error`` take effect when the program exits with an exception, including the
    KeyboardInterrupt caused by CTRL-C.

    ``…_on_reload`` take effect when the program is interrupted by files being written to the USB
    drive (auto-reload) or when it calls ``supervisor.reload()``.

    These settings are stored in RAM, not in persistent memory, and will therefore only affect
    soft reloads. Powering off or resetting the device will always revert to standard settings.

    When called multiple times in the same run, only the last call takes effect, replacing any
    settings made by previous ones. This is the main use of passing ``None`` as a filename: to
    reset to the standard search sequence."""
    ...

class RunReason:
    """The reason that CircuitPython started running."""

    STARTUP: object
    """CircuitPython started the microcontroller started up. See `microcontroller.Processor.reset_reason`
       for more detail on why the microcontroller was started."""

    AUTO_RELOAD: object
    """CircuitPython restarted due to an external write to the filesystem."""

    SUPERVISOR_RELOAD: object
    """CircuitPython restarted due to a call to `supervisor.reload()`."""

    REPL_RELOAD: object
    """CircuitPython started due to the user typing CTRL-D in the REPL."""

class Runtime:
    """Current status of runtime objects.

    Usage::

       import supervisor
       if supervisor.runtime.serial_connected:
           print("Hello World!")"""

    def __init__(self) -> None:
        """You cannot create an instance of `supervisor.Runtime`.
        Use `supervisor.runtime` to access the sole instance available."""
        ...
    usb_connected: bool
    """Returns the USB enumeration status (read-only)."""

    serial_connected: bool
    """Returns the USB serial communication status (read-only)."""

    serial_bytes_available: int
    """Returns the whether any bytes are available to read
    on the USB serial input.  Allows for polling to see whether
    to call the built-in input() or wait. (read-only)"""

    run_reason: RunReason
    """Returns why CircuitPython started running this particular time."""



# user defined iterator used in something other than a for loop

class MyStopIteration(StopIteration):
    pass

class myiter:
    def __init__(self, i):
        self.i = i

    def __iter__(self):
        return self

    def __next__(self):
        if self.i == 0:
            raise StopIteration
        elif self.i == 1:
            raise StopIteration(1)
        elif self.i == 2:
            raise MyStopIteration

print(list(myiter(0)))
print(list(myiter(1)))
print(list(myiter(2)))



# test floating point floor divide and modulus
# it has some tricky corner cases

# pyboard has 32-bit floating point and gives different (but still
# correct) answers for certain combinations of divmod arguments.


def test(x, y):
    div, mod = divmod(x, y)
    print(div == x // y, mod == x % y, abs(div * y + mod - x) < 1e-6)


test(1.23456, 0.7)
test(-1.23456, 0.7)
test(1.23456, -0.7)
test(-1.23456, -0.7)

a = 1.23456
b = 0.7
test(a, b)
test(a, -b)
test(-a, b)
test(-a, -b)

for i in range(25):
    x = (i - 12.5) / 6
    for j in range(25):
        y = (j - 12.5) / 6
        test(x, y)

# test division by zero error
try:
    divmod(1.0, 0)
except ZeroDivisionError:
    print("ZeroDivisionError")
